/*
 * OBServ TS Services, Copyright (C) 1998 Patrick Alken
 * This program comes with absolutely NO WARRANTY
 *
 * Should you choose to use and/or modify this source code, please
 * do so under the terms of the GNU General Public License under which
 * this program is distributed.  You know the drill.
 */

#include	<stdio.h>
#include	<string.h>
#include	<sys/types.h>
#include 	<netinet/in.h>
#include	<arpa/inet.h>
#include	<netdb.h>
#include	"observ.h"
#include	"config.h"

extern		char	AdminInfo[MAXLINE];
extern		char	*version;
extern		int		errno;

/*
BroadcastDCC()
	args: int cindex, char *msg, int towho
	purpose: send 'msg' to all DCC CHAT users specified by 'towho'
	return: none
*/

void BroadcastDCC(int cindex, char *msg, int towho)

{
	int		ii;
	char	sendstr[MAXLINE];

	for (ii = 0; ii < MAXDCCCONNS; ii++)
	{
		if (connections[ii].socket != -1)
			switch (towho)
			{
				case DCCALL:
				{
					if (cindex == -1) 
					{
						/* it's a general message for user joining partyline etc */
						if (!connections[ii].isbot)
							writesocket(connections[ii].socket, msg);
						else
						{
							sprintf(sendstr, "(%s) %s",
								SERVICENICK,
								msg);
							writesocket(connections[ii].socket, sendstr);
						}
					}
					else /* it's a specific message from a client */
					{
						sprintf(sendstr, "<%s@%s> %s\n",
							connections[cindex].nick,
							SERVICENICK,
							msg);
						writesocket(connections[ii].socket, sendstr);
					}
					break;
				}
				case DCCOPS:
				{
					int	userindex;

					userindex = GetUser(connections[ii].nick);
					if (IsRegistered(connections[ii].nick) == 1)
						if ((CheckAccess(userindex, "o") == 1) || 
								(CheckAccess(userindex, "a") == 1))
						{
							sprintf(sendstr, "<%s@%s> %s\n",
								connections[cindex].nick,
								SERVICENICK,
								msg);
							writesocket(connections[ii].socket, msg);
						}
					break;
				}
			} /* switch (towho) */
	}
} /* BroadcastDCC() */

/*
MakeConn()
	args: char *host, int port, char *nick, char *userhost
	purpose: create a DCC CHAT session for 'nick'
	return: 1 if connection was successful, 0 if not
*/

int MakeConn(char *host, int port, char *nick, char *userhost)

{
	int		ii, userindex, numconns;
	char	dccbuff[MAXBUFF];
	char	sendstr[MAXLINE], prefix[MAXLINE];

	numconns = 0;
	for (ii = 0; ii <= MAXDCCCONNS; ii++)
	{
		if (connections[ii].socket != -1)
			numconns++;
	}
	if (numconns > MAXDCCCONNS) /* maximum users reached */
	{
		sprintf(sendstr, "Maximum DCC connections have been reached, try again later");
		notice(nick, sendstr, NODCC);
		return 0;
	}

	for (ii = 0; ii < MAXDCCCONNS; ii++)
		if (connections[ii].socket == -1)
			break;

	connections[ii].socket = ConnectToServer(host, port);
	if (connections[ii].socket == -1)
	{
		notice(nick, "DCC CHAT connection failed", NODCC);
		return 0;
	}

	strncpy(connections[ii].nick, nick, MAXLINE - 1);
	connections[ii].nick[MAXLINE - 1] = '\0';
	strncpy(connections[ii].userhost, userhost, MAXLINE - 1);
	connections[ii].userhost[MAXLINE - 1] = '\0';

	SendMOTD(connections[ii].socket); /* give new user the motd */

	if (IsRegistered(connections[ii].nick) != 1)
		writesocket(connections[ii].socket, 
			"You must .register before you may use commands\r\n\r\n");

	userindex = GetUser(connections[ii].nick);
	if (CheckAccess(userindex, "a") == 1)
		strcpy(prefix, "Admin");
	else
		if (CheckAccess(userindex, "o") == 1)
			strcpy(prefix, "Oper");
		else
			strcpy(prefix, "User");

	sprintf(sendstr, "%s %s (%s) has connected\n",
		prefix,
		connections[ii].nick,
		connections[ii].userhost);
	BroadcastDCC(-1, sendstr, DCCALL); /* inform everyone of new user */

	return 1;
} /* MakeConn() */

/*
onctcp()
	args: char *nick, char *msg
	purpose: handle ctcp message 'msg' sent by 'nick'
	return: none
*/

void onctcp(char *nick, char *msg)

{
	struct	sockaddr_in	ip_addr;
	char	buff[MAXBUFF];
	char	port[MAXBUFF];
	char	temp[MAXLINE];
	char	userhost[MAXLINE];
	int		luindex, uindex;
	int		goodDCC;
	int		ii;

	buff[0] = '#';
	++msg;
	luindex = IsUser(nick);
	if (strncasecmp(msg, "PING", 4) == 0)
	{
		notice(nick, msg - 1, NODCC);
		putlog(LOG1, "(ctcp!ping) recieved from %s!%s",
			Lusers[luindex].Nick,
			Lusers[luindex].Userhost);
		return;
	}
	if (strncasecmp(msg, "VERSION", 7) == 0)
	{
		sprintf(buff, "\001VERSION OBServ version %s\001", version); 
		notice(nick, buff, NODCC);
		sprintf(buff, "\001VERSION Administrators: %s\001", AdminInfo);
		notice(nick, buff, NODCC);
		putlog(LOG1, "(ctcp!version) recieved from %s!%s",
			Lusers[luindex].Nick,
			Lusers[luindex].Userhost);
		return;
	}
	if (strncasecmp(msg, "DCC CHAT", 8) == 0)
	{
		goodDCC = 1;
		if (luindex == -1)
		{
			for (ii = 0; ii < MAXLUSERS; ii++)
				if (strcasecmp(Lusers[ii].Nick, nick) == 0)
				{
					luindex = ii;
					break;
				}
			goodDCC = 0;
		}

		strcpy(userhost, Lusers[luindex].Userhost);
		for (ii = 1; ii < MAXBUFF; ii++)
			buff[ii] = '\0';
		GetStrIndex(temp, msg, 4);
		uindex = GetUser(nick);
		if (uindex != -1)
		{
			if ((CheckAccess(uindex, "d") == 1) ||
					(CheckAccess(uindex, "o") == 1))
			{
				goodDCC = 1;
			#ifdef OPERSONLYDCC
				if (CheckAccess(uindex, "o") != 1)
					goodDCC = 0;
			#endif
			}
			else
				goodDCC = 0;
			if (CheckAccess(uindex, "a") == 1)
				goodDCC = 1;
		} /* if (uindex != -1) */
		if (goodDCC == 0)
		{
			notice(nick, "You are not authorized to use this service", NODCC);
			sprintf(temp, "Unauthorized DCC CHAT attempt from %s (%s)\n",
				nick,
				userhost);
			BroadcastDCC(-1, temp, DCCOPS);
			notice(nick, "\001DCC REJECT chat chat\001", NODCC);
			putlog(LOG1, "Unauthorized DCC CHAT attempt from %s!%s", 
				nick,
				userhost);
			return;
		}

		strcat(buff, temp);
		GetStrIndex(port, msg, 5);
		if (atoi(port) < 1024)
		{
			notice(nick, "Invalid DCC CHAT port specified.  Nice Try.", NODCC);
			notice(nick, "\001DCC REJECT chat chat\001", NODCC);
			putlog(LOG1, "Bogus DCC CHAT port specified by %s!%s",
				nick,
				userhost);
		}
		else 
		{
			if (!MakeConn(buff, atoi(port), nick, userhost))
			{
				notice(nick, "DCC CHAT connection failed", NODCC);
				notice(nick, "\001DCC REJECT chat chat\001", NODCC);
			}
			return;
		}
	}
	else
	{
		strncpy(temp, msg, strlen(msg) - 1); /* chop off ending \001 */
		putlog(LOG1, "(ctcp!%s) recieved from %s!%s",
			temp,
			Lusers[luindex].Nick,
			Lusers[luindex].Userhost);
	}
} /* onctcp() */

/*
ConnectToTCM()
	args: char *nick, int bindex
	purpose: attempt to connect to tcm bot 'bot[bindex].BotName'
	return: 1 on a successful connection, 0 if not
*/

int ConnectToTCM(char *nick, int bindex)

{
	int		ii, jj, nindex, lindex;
	char	sendstr[MAXLINE];

	for (ii = 0; ii < MAXDCCCONNS; ii++)
		if (connections[ii].socket == -1)
			break;

	nindex = IsOnDCC(nick);
	connections[ii].socket = ConnectToServer(bots[bindex].BotHost, 
																					 atoi(bots[bindex].BotPort));

	if (connections[ii].socket == -1)
	{
		sprintf(sendstr, "Unable to connect to port %s of %s: %s\n", 
			bots[bindex].BotPort,
			bots[bindex].BotHost,
			strerror(errno));
		writesocket(connections[nindex].socket, sendstr);
		return (0);
	}

	/* 
		The fields for linking to a tcm are in order:
		<remote tcm's nick> <connecting bot's nick> <password>
	*/
	sprintf(sendstr, "%s %s %s\n",
		bots[bindex].BotName, 
		SERVICENICK,
		bots[bindex].BotPass);
	writesocket(connections[ii].socket, sendstr);

	connections[ii].isbot = 1;
	connections[ii].ishub = 1;
	connections[ii].ispending = 0;
	strcpy(connections[ii].nick, bots[bindex].BotName);

	lindex = IsLuser(connections[ii].nick); /* check if bot is on network */
	if (lindex == -1) /* just make the bot's hostname ip:port */
	{
		sprintf(sendstr, "%s:%s",
			bots[bindex].BotHost,
			bots[bindex].BotPort);
		strcpy(connections[ii].userhost, sendstr);
	}
	else /* give the bot a user@host if possible */
		strcpy(connections[ii].userhost, Lusers[lindex].Userhost);

	for (jj = 0; jj < MAXDCCCONNS; jj++) /* inform everyone of new bot */
	{
		if (jj == ii)
			continue;
		if (connections[jj].isbot)
		{
			sprintf(sendstr, "(%s) Linked to %s [%s]\n",
				SERVICENICK,
				bots[bindex].BotName,
				connections[ii].userhost);
			writesocket(connections[jj].socket, sendstr);
		}
		else
		{
			sprintf(sendstr, "*** Linked to %s [%s]\n",
				bots[bindex].BotName,
				connections[ii].userhost);
			writesocket(connections[jj].socket, sendstr);
		}
	} 

	return(1);
} /* ConnectToTCM */

/*
CloseConnection()
	args: int cindex
	purpose: close socket of DCC CHAT client specified by 'cindex'
	return: none
*/

void CloseConnection(int cindex)

{
	int		userindex;
	char	prefix[MAXLINE], sendstr[MAXLINE];

	if (connections[cindex].socket != -1)
		close(connections[cindex].socket);

	connections[cindex].socket = -1;

	if (connections[cindex].isbot == 1)
		strcpy(prefix, "Bot");
	else
	{
		userindex = GetUser(connections[cindex].nick);
		if (CheckAccess(userindex, "a") == 1)
			strcpy(prefix, "Admin");
		else
			if (CheckAccess(userindex, "o") == 1)
				strcpy(prefix, "Oper");
			else
				strcpy(prefix, "User");
	}

	sprintf(sendstr, "%s %s (%s) has disconnected\n",
		prefix,
		connections[cindex].nick,
		connections[cindex].userhost);
	BroadcastDCC(-1, sendstr, DCCALL);

	connections[cindex].nick[0] = '\0';
	connections[cindex].userhost[0] = '\0';
} /* CloseConnection() */

/*
ExecuteDCCCommand()
	args: int cindex, int cmdtok, char *args
	purpose: execute command (cmdtok) sent by connections[cindex].nick
	return: none
*/

void ExecuteDCCCommand(int cindex, int cmdtok, char *args)

{
	int		ii;
	int		userindex;
	char	sendstr[MAXLINE];
	char	nargs[MAXLINE];

	userindex = GetUser(connections[cindex].nick);

	for (ii = 0; ii < MAXLINE; ii++)
	{
		sendstr[ii] = '\0';
		nargs[ii] = '\0';
	}
	if (args == (char *) NULL)
		strcpy(nargs, "");
	else
		strcpy(nargs, args);

	switch (cmdtok)
	{
		case D_HELP:
		{
			GiveHelp(NULL, nargs, connections[cindex].socket);
			sprintf(sendstr, "#%s# help %s\n",
				connections[cindex].nick,
				nargs);
			break;
		} /* D_HELP */
		case D_WHO:
		{
			int 	uindex;
			int		AreBots = NO;
			char	prefix[MAXLINE];

			writesocket(connections[cindex].socket, 
				"Current users: (* = Admin, + = Oper)\n");
			for (ii = 0; ii < MAXDCCCONNS; ii++)
				if (connections[ii].socket != -1)
				{
					if (!connections[ii].isbot)
					{
						uindex = GetUser(connections[ii].nick);
						if (CheckAccess(uindex, "a") == 1)
							strcpy(prefix, "*");
						else
							if (CheckAccess(uindex, "o") == 1)
								strcpy(prefix, "+");
							else
								strcpy(prefix, " ");
						sprintf(sendstr, "  %s%s [%s]\n",
							prefix,
							connections[ii].nick,
							connections[ii].userhost);
						writesocket(connections[cindex].socket, sendstr);
					}
					else
						AreBots = YES;
				}
			if (AreBots == YES)
			{
				writesocket(connections[cindex].socket, "\r\nLinked Bots:\n");
				for (ii = 0; ii < MAXDCCCONNS; ii++)
				{
					if (connections[ii].socket != -1)
					{
						if (connections[ii].isbot)
						{
							if (connections[ii].ishub)
								sprintf(sendstr, "  -> %s [%s]\n",
									connections[ii].nick,
									connections[ii].userhost);
							else
								sprintf(sendstr, "  <- %s [%s]\n",
									connections[ii].nick,
									connections[ii].userhost);
							writesocket(connections[cindex].socket, sendstr);
						}
					}
				}
			}
			sprintf(sendstr, "#%s# who\n", connections[cindex].nick);
			break;
		} /* D_WHO */
		case D_BOOT: /* forcefully remove someone from partyline */
		{
			char	bootnick[MAXLINE];
			char	reason[MAXLINE];
			char	temp[MAXLINE];
			int		uindex;

			GetStrIndex(bootnick, nargs, 1);
			GetStrLRange(reason, nargs, 2);
			if (CheckAccess(userindex, "a") == 1)
			{
				for (ii = 0; ii < MAXDCCCONNS; ii++)
				{
					if (connections[ii].socket == -1)
						continue;
					if (!strcasecmp(connections[ii].nick, bootnick))
					{
						uindex = GetUser(connections[ii].nick);
						if ((CheckAccess(uindex, "a") == 1) &&
								(IsRegistered(connections[ii].nick) == 1))
						{
							writesocket(connections[cindex].socket, "Cannot boot an administrator\n");
							break;
						}
						sprintf(sendstr, "You have been booted by %s [%s]\n",
							connections[cindex].nick,
							reason);
						writesocket(connections[ii].socket, sendstr);
						close(connections[ii].socket);
						connections[ii].socket = -1;
						sprintf(sendstr, "%s has been booted by %s [%s]\n",
							connections[ii].nick,
							connections[cindex].nick,
							reason);
						BroadcastDCC(-1, sendstr, DCCALL);
						connections[ii].nick[0] = '\0';
						connections[ii].userhost[0] = '\0';
						sprintf(sendstr, "#%s# boot %s %s\n", 
							connections[cindex].nick,
							bootnick,
							reason);
						break;
					} /* if (!strcasecmp(connections[ii].nick, bootnick)) */
				} /* for (ii = 0; ii < MAXDCCCONNS; ii++) */
			} /* if (CheckAccess(userindex, "a") == 1) */
			else
				writesocket(connections[cindex].socket, "No privileges\n");
			break;
		} /* D_BOOT */
		case D_QUIT:
		{
			sprintf(sendstr, "Closing connection: %s\n", nargs);
			writesocket(connections[cindex].socket, sendstr);
			close(connections[cindex].socket);
			connections[cindex].socket = -1;
			sprintf(sendstr, "%s (%s) has left the partyline [%s]\n",
				connections[cindex].nick,
				connections[cindex].userhost,
				nargs);
			BroadcastDCC(-1, sendstr, DCCALL);
			connections[cindex].nick[0] = '\0';
			connections[cindex].userhost[0] = '\0';
			sendstr[0] = '\0';
			break;
		} /* D_QUIT */
		case D_MOTD:
		{
			SendMOTD(connections[cindex].socket);
			sprintf(sendstr, "#%s# motd\n", connections[cindex].nick);
			break;
		} /* D_MOTD */
		case D_LINK:
		{
			char	botname[MAXLINE];
			int		bindex;

			GetStrIndex(botname, nargs, 1);
			if ((CheckAccess(userindex, "a") == 1) ||
					(CheckAccess(userindex, "o") == 1))
			{
				if ((bindex = IsBot(botname)) == -1)
				{
					sprintf(sendstr, "No such bot: %s\n", botname);
					writesocket(connections[cindex].socket, sendstr);
				}
				else
				{
					sprintf(sendstr, "Attempting to link to %s\n", 
						bots[bindex].BotName);
					writesocket(connections[cindex].socket, sendstr);
					sprintf(sendstr, "Making connection to %s:%s\n",
						bots[bindex].BotHost,
						bots[bindex].BotPort);
					writesocket(connections[cindex].socket, sendstr);
					ConnectToTCM(connections[cindex].nick, bindex);
				}
				sprintf(sendstr, "#%s# link %s\n", 
					connections[cindex].nick,
					botname);
			}
			else
				writesocket(connections[cindex].socket, "No privileges\n");
			break;
		} /* D_LINK */
		case D_UNLINK:
		{
			char 	botname[MAXLINE];
			char	reason[MAXLINE];
			char	origreason[MAXLINE];
			int		bindex;

			if ((CheckAccess(userindex, "a") == 1) ||
					(CheckAccess(userindex, "o") == 1))
			{
				GetStrIndex(botname, nargs, 1);
				GetStrLRange(reason, nargs, 2);
				if (botname[0] == '\0')
				{
					writesocket(connections[cindex].socket, 
						"Usage: .unlink <bot name> [reason]\n");
					return;
				}
				bindex = GetBot(botname);
				if (bindex == -1)
				{
					sprintf(sendstr, "I am not linked to %s\n",
						botname);
					writesocket(connections[cindex].socket, sendstr);
					return;
				}
				if (reason[0] == '\0')
				{
					origreason[0] = '\0';
					sprintf(sendstr, "(%s)",
						connections[cindex].nick);
					strcpy(reason, sendstr);
				}
				else
				{
					strcpy(origreason, reason);
					sprintf(sendstr, "(%s: %s)",
						connections[cindex].nick,
						reason);
					strcpy(reason, sendstr);
				}
				if (connections[bindex].socket != -1)
				{
					close(connections[bindex].socket);
					connections[bindex].socket = -1;
					connections[bindex].userhost[0] = '\0';
					connections[bindex].isbot = 0;
					connections[bindex].ishub = 0;
					connections[bindex].ispending = 1;
				}
				sprintf(sendstr, "*** Unlinked from %s %s\n",
					connections[bindex].nick,
					reason);
				BroadcastDCC(-1, sendstr, DCCALL);
				connections[bindex].nick[0] = '\0';
				sprintf(sendstr, "#%s# unlink %s %s\n",
					connections[cindex].nick,
					botname,
					origreason);
			}
			else
				writesocket(connections[cindex].socket, "No privileges\n");
			break;
		} /* D_UNLINK */
	} /* switch (cmdtok) */

	/* notify ops/admins of the command the user used */
	if (sendstr[0] != '\0')
		BroadcastDCC(-1, sendstr, DCCOPS); 
} /* ExecuteDCCCommand() */

/*
DCCProcess()
	args: int cindex, char *command
	purpose: process 'command' sent by DCC client specified by 'cindex'
	return: none
*/

void DCCProcess(int cindex, char *command)

{
	char	buff[MAXBUFF];
	char	*args;
	char	keyword[MAXLINE];
	int		token;
	int		ii;

	if (*command == '.')
	{
		GetStrRange(buff, command, 1, strlen(command));
		GetStrIndex(keyword, buff, 1);
		if (IsRegistered(connections[cindex].nick) == 1)
		{
			token = 0;
			if (strncasecmp(keyword, "HELP", 4) == 0)
				token = D_HELP;	
			if (strncasecmp(keyword, "WHO", 3) == 0)
				token = D_WHO;
			if (strncasecmp(keyword, "BOOT", 4) == 0)
				token = D_BOOT;
			if (strncasecmp(keyword, "QUIT", 4) == 0)
				token = D_QUIT;
			if (strncasecmp(keyword, "MOTD", 4) == 0)
				token = D_MOTD;
			if (strncasecmp(keyword, "LINK", 4) == 0)
				token = D_LINK;
			if (strncasecmp(keyword, "UNLINK", 6) == 0)
				token = D_UNLINK;

			if (token == 0)
				ProcessCommand(connections[cindex].nick, buff, connections[cindex].socket);
			else
			{
				args = strtok(buff, " "); 
				args = strtok((char *) NULL, "\r\n");
				ExecuteDCCCommand(cindex, token, args);
			}
			return;
		}
		if (strncasecmp(keyword, "REGISTER", 8) == 0) 
			ProcessCommand(connections[cindex].nick, buff, connections[cindex].socket);
		else
			if (strncasecmp(keyword, "HELP", 4) == 0)
			{
				args = strtok(buff, " ");
				args = strtok((char *) NULL, "\r\n");
				ExecuteDCCCommand(cindex, D_HELP, args);
			}
			else	
				writesocket(connections[cindex].socket, "You must register first\n");
	}
	else /* it was not a command, broadcast it to dcc users */
	{
		BroadcastDCC(cindex, command, DCCALL);
	}
} /* DCCProcess() */

/*
BotProcess()
	args: int bindex, char *line
	purpose: send 'line' to all other bots/users on botnet coming from
					 bots[bindex].BotName
	return: none
*/

void BotProcess(int bindex, char *line)

{
	int		ii;

	if (connections[bindex].ispending)
	{
		char	*mynick;
		char	*tcmnick;
		char	*password;
		int		aindex;
		char	sendstr[MAXLINE];

		mynick = strtok(line, " ");
		tcmnick = strtok((char *) NULL, " ");
		password = strtok((char *) NULL, "\r\n");

		printf("stuff=%s %s %s\n", mynick, tcmnick, password);

		aindex = GoodTCM(connections[bindex].userhost);
		printf("a=%d %s %s\n", aindex, AllowedBots[aindex].BotName,
		AllowedBots[aindex].BotHost);

		if ((mynick == (char *) NULL) || 
				(strcasecmp(mynick, SERVICENICK) != 0))
		{
			sprintf(sendstr, "Invalid myname on connection attempt from %s\n",
				connections[bindex].userhost);
			BroadcastDCC(-1, sendstr, DCCOPS);
			putlog(LOG1, sendstr);
			close(connections[bindex].socket);
			connections[bindex].socket = -1;
			return;
		}
	
		if ((tcmnick == (char *) NULL) ||
				(strcasecmp(tcmnick, AllowedBots[aindex].BotName) != 0))
		{
			sprintf(sendstr, "Invalid bot name [%s] specified on connection attempt from %s\n",
				tcmnick,
				connections[bindex].userhost);
			BroadcastDCC(-1, sendstr, DCCOPS);
			putlog(LOG1, sendstr);
			close(connections[bindex].socket);
			connections[bindex].socket = -1;
			return;
		}

		if ((password == (char *) NULL) ||
				(strcasecmp(password, AllowedBots[aindex].BotPass) != 0))
		{
			sprintf(sendstr, "Invalid password specified on connection attempt from %s\n",
				connections[bindex].userhost);
			BroadcastDCC(-1, sendstr, DCCOPS);
			putlog(LOG1, sendstr);
			close(connections[bindex].socket);
			connections[bindex].socket = -1;
			return;
		}

		strcpy(connections[bindex].nick, tcmnick);
		connections[bindex].ispending = 0;
		sprintf(sendstr, "Connection with %s [%s] established\n",
			connections[bindex].nick,
			connections[bindex].userhost);
		BroadcastDCC(-1, sendstr, DCCOPS);
		putlog(LOG1, sendstr);
		sprintf(sendstr, "*** Linked to %s [%s]\n",
			connections[bindex].nick,
			connections[bindex].userhost);
		BroadcastDCC(-1, sendstr, DCCALL);

		return;
	}

	if (line[0] == '<') /* <nick@bot> said something */
	{
		for (ii = 0; ii < MAXDCCCONNS; ii++)
			if (connections[ii].socket != -1)
			{
				if (bindex == ii)
					continue; /* don't send it back to the bot it came from */
				writesocket(connections[ii].socket, line);
			}
	}
} /* BotProcess() */
