/*
 * OBServ TS Services, Copyright (C) 1998 Patrick Alken
 * This program comes with absolutely NO WARRANTY
 *
 * Should you choose to use and/or modify this source code, please
 * do so under the terms of the GNU General Public License under which
 * this program is distributed.  You know the drill.
 */

#include	<stdio.h>
#include	<sys/types.h>
#include	<netdb.h>
#include	<sys/socket.h>
#include	<netinet/in.h>
#include 	<arpa/inet.h>
#include	<sys/errno.h>
#include	<sys/time.h>
#include	<string.h>
#include	<varargs.h>
#include	<signal.h>
#include	"config.h"
#include	"observ.h"

char			*version = "1.1.3";
extern		int		errno;
int				CurrSIndex; /* Current index of server being used */
fd_set		readfds, nullfds;
extern		char	MyName[MAXLINE];
char			ConfigFile[MAXFILELENGTH]; /* name of config file */
extern		struct	DCCStruct connections[];
int				TcmSocket; /* socket to listen on for tcm bots */

/*
writesocket()
	args: sockfd, writestr
	purpose: write 'writestr' to socket designated by 'sockfd'
	return: none
*/

void writesocket(int sockfd, char *writestr)
{
	send(sockfd, writestr, strlen(writestr), 0);
}

/*
putlog()
	args: int level, char *fmt, va_alist
	purpose: log 'fmt' to log file
	return: none
*/

void putlog(level, fmt, va_alist)
int level;
char *fmt;
va_dcl

{
	FILE	*fp;
	time_t	CurrTime;
	char	putstr[MAXLINE];
	va_list	ap;

	if (LOGLEVEL == LOG0)
		return; /* LOG0 means don't log anything */

	if (LOGLEVEL < level)
		return;

	va_start(ap);
	if ((fp = fopen(LOGFILE, "a+")) == (FILE *) NULL)
	{
	#ifdef DEBUGMODE
		printf("Unable to open log file: %s\n", LOGFILE);
	#endif
		return;
	}
	CurrTime = time(NULL);
	strcpy(putstr, ctime(&CurrTime));

	/* Erase the \n character that ctime() puts on the end, in order to
		 concat the string we want to log */
	putstr[strlen(putstr) - 1] = ' '; 

	strcat(putstr, fmt);
	strcat(putstr, "\n");
	vfprintf(fp, putstr, ap); /* write the string to the log file */

	#ifdef DEBUGMODE
		sprintf(putstr, "Logged> ");
		strcat(putstr, fmt);
		strcat(putstr, "\n");
		vprintf(putstr, ap);
	#endif

	fclose(fp);
	va_end(ap);
} /* putlog() */

/*
ConnectToServer()
	args: hostname, port
	purpose: bind socket and establish connection to hostname at port
	return: socket file descriptor if successful connect, -1 if not
*/

int	ConnectToServer(char *hostname, int port)
{
	struct  sockaddr_in	serv_addr;
	char    ServHostIP[25]; /* Internet address for server hostname */
	char		host[MAXLINE];
	register struct hostent	*hostptr;
	struct 	in_addr	*ptr;
	int			socketfd; /* socket file descriptor */

	strcpy(host, hostname);
	bzero((char *) &serv_addr, sizeof(serv_addr));
	if (*hostname == '#') /* DCC CHAT host */
	{
		serv_addr.sin_addr.s_addr = htonl(strtoul(hostname + 1,(char **) NULL, 10));
#ifdef DEBUGMODE
		printf("Connecting to port %d of %s (DCC CHAT)\n", port, inet_ntoa(serv_addr.sin_addr));
#endif
		strcpy(host, inet_ntoa(serv_addr.sin_addr));
		putlog(LOG2, "Attempting to connect to %s:%d (DCC CHAT)", inet_ntoa(serv_addr.sin_addr), port); 
	}
	else
	{
		if ((hostptr = gethostbyname(hostname)) == NULL)
		{
		#ifdef DEBUGMODE
			printf("Cannot connect to port %d of %s: Unknown host\n", port, hostname);
		#endif
			return(-1);
		}
		else
		{
			ptr = (struct in_addr *) *hostptr->h_addr_list;
			strcpy(ServHostIP, inet_ntoa(*ptr));
			serv_addr.sin_addr.s_addr = inet_addr(ServHostIP);
		#ifdef DEBUGMODE
			printf("Connecting to port %d of %s (%s)\n", port, hostptr->h_name, ServHostIP); 
		#endif
			putlog(LOG1, "Attempting to connect to %s:%d", hostptr->h_name, port);
		}
	}
	serv_addr.sin_family = AF_INET;
	
	serv_addr.sin_port = htons(port);
	if((socketfd = socket(AF_INET, SOCK_STREAM, 0)) < 0)
	{
	#ifdef DEBUGMODE
		printf("Unable to open stream socket\n");
	#endif
	}

	while (connect(socketfd, (struct sockaddr *) &serv_addr, sizeof(serv_addr)) < 0)
		if (errno != EINTR)
		{
#ifdef DEBUGMODE
			printf("Cannot connect to port %d of %s: %s\n", port, host, strerror(errno));
#endif
			return(-1);
		}
#ifdef DEBUGMODE
	printf("Connection established\n");
#endif
	putlog(LOG1, "Successfully connected to %s", host);
	return(socketfd);
} /* ConnectToServer() */

/*
ProcessInfo()
	args: char *info
	purpose: process string sent by hub server (stored in 'info')
	return: none
*/
void ProcessInfo(char *info)
{
	char	secondword[MAXLINE];
	int		token;
	int		luserindex;
	char	nick[MAXLINE]; 
	char	temp[MAXLINE];

#ifdef DEBUGMODE
	printf("%s", info);
#endif
	GetStrIndex(temp, info, 1);
	GetStrRange(nick, temp, 1, strlen(temp));
	token = 0;
	if (strncasecmp(info, "PASS", 4) == 0)
		token = S_PASS;
	if (strncasecmp(info, "PING", 4) == 0)
		token = S_PING;
	if (strncasecmp(info, "SERVER", 6) == 0)
		token = S_SERVER;
	if (strncasecmp(info, "NICK", 4) == 0)
		token = S_NICK;

	/* This switch is for the commands such as PING, SERVER, PASS, */
	/* NICK, etc, which come first.  Another switch will be used   */
	/* later for commands which come second, like PRIVMSG          */

	switch (token) 
	{							 
		case S_PASS:
			InitServs(); /* Since the PASS line only occurs once */
			break; 			 /* introduce the servs here, so they    */
									 /* will only be introduced once         */
		case S_PING:
		{
			char	putstr[MAXLINE];
			char	ServName[MAXLINE];

			GetStrIndex(ServName, info, 2);
			(void)sprintf(putstr, ":%s PONG %s %s\n", MyName, MyName, 
										ServName);
			writesocket(Servers[CurrSIndex].socket, putstr);
			break;
		}
		case S_SERVER: /* get the network server name of hub in case it */
									 /* doesn't match it's hostname									  */
		{
			GetStrIndex(Servers[CurrSIndex].RealName, info, 2);
			break;
		} /* S_SERVER */
		case S_NICK:
		{
			char	userhost[MAXLINE];
			char	nick[MAXLINE];
			char	user[MAXLINE], host[MAXLINE];

			GetStrIndex(user, info, 6);
			GetStrIndex(host, info, 7);
			strcpy(userhost, strcat(strcat(user, "@"), host));
			GetStrIndex(nick, info, 2);
			
			/* Add all users on network, to be able to check the user@host */
			/* for any of them on PRIVMSG to see if they have access			 */
			AddUserInfo(nick, userhost); 

			CheckGlined(nick, userhost);
			break;
		}

		default:
			break;
	}
	if (token != 0) /* If token doesn't equal 0 anymore, then the      */
		return;				/* command was taken care of in the above switch() */ 
									/* so, might as well not go through this one :-)   */

	GetStrIndex(secondword, info, 2);
	if (strncasecmp(secondword, "PRIVMSG", 7) == 0)
		token = U_PRIVMSG;
	if (strncasecmp(secondword, "QUIT", 4) == 0)
		token = U_QUIT;
	if (strncasecmp(secondword, "NICK", 4) == 0)
		token = U_NICK;	
	if (strncasecmp(secondword, "SERVER", 6) == 0)
		token = U_SERVER;
	if (strncasecmp(secondword, "353", 3) == 0) /* /names reply */
		token = U_RPLNAMES;	
	if (strncasecmp(secondword, "366", 3) == 0) /* end of names */
		token = U_EONAMES;
	if (strncasecmp(secondword, "324", 3) == 0)	/* MODE reply */
		token = U_RPLMODE;
	if (strncasecmp(secondword, "KILL", 4) == 0)
		token = U_KILL;
	if (strncasecmp(secondword, "KICK", 4) == 0)
		token = U_KICK;
	if (strncasecmp(secondword, "MODE", 4) == 0)
		token = U_MODE;
	if (strncasecmp(secondword, "367", 3) == 0)
		token = U_RPLBAN;
	
	luserindex = IsLuser(nick);

	switch (token)
	{
		case U_PRIVMSG:
		{
			char	tempinfo[MAXLINE]; /* for strcpy() purposes */
			char	temp2[MAXLINE]; /* for strcpy() purposes */
			char	command[MAXLINE]; /* contains the /msg that 'nick' gave */
			char	sendstr[MAXBUFF]; /* for writesocket() */
			char	dest[MAXLINE], snick[MAXLINE];
			int		ii; /* looping */

			for (ii = 0; ii < MAXLINE; ii++)
			{
				tempinfo[ii] = '\0';
				temp2[ii] = '\0';
				command[ii] = '\0';
				dest[ii] = '\0';
				snick[ii] = '\0';
			}
			GetStrIndex(dest, info, 3);
			if (strncasecmp(dest, SERVICENICK, strlen(SERVICENICK)) != 0)
				return; /* The PRIVMSG was not for OBServ */
			GetStrLRange(temp2, info, 4);

			/* remove the leading ":" from the command string */
			GetStrRange(command, temp2, 1, strlen(temp2)); 

			for (ii = 0; ii < MAXLINE; ii++)
				temp2[ii] = '\0';
			if (command[0] == '\001')
			{
				onctcp(nick, command);
				return;
			}

			if (IsRegistered(nick) == YES)
				ProcessCommand(nick, command, NODCC);
			else
			{
				if (GetUser(nick) >= 0)
				{
					GetStrIndex(temp2, command, 1);
					if ((strncasecmp(temp2, "REGISTER", 8) == 0) || 
							(strncasecmp(temp2, "HELP", 4) == 0))
						ProcessCommand(nick, command, NODCC);
					else
						notice(nick, "You must register first", NODCC);
				}
				else
					notice(nick, "Access Denied", NODCC);
			}
			break;		
		} /* U_PRIVMSG */
		case U_QUIT:
		{
			GetStrIndex(temp, info, 1);
			GetStrRange(nick, temp, 1, strlen(temp));
			RemUserInfo(nick); /* delete user when they quit */
			break;
		} /* U_QUIT */
		case U_NICK: /* Nick change */
		{
			char	newnick[MAXLINE];
			char	sendstr[MAXLINE];
			int		jj;

			GetStrIndex(temp, info, 1);
			GetStrRange(nick, temp, 1, strlen(temp));
			luserindex = IsLuser(nick);
			GetStrIndex(newnick, info, 3);

			/* send nick change to partyline if nick is on it */
			if (IsOnDCC(Lusers[luserindex].Nick) != -1)
			{
				strcpy(connections[IsOnDCC(Lusers[luserindex].Nick)].nick, newnick);
				sprintf(sendstr, "Nick change: %s -> %s\n", 
					Lusers[luserindex].Nick,
					newnick);
				BroadcastDCC(-1, sendstr, DCCALL);
			}
			
			/* Erase old nick, and store new nick in it's position */
			strcpy(Lusers[luserindex].Nick, newnick);

			break;
		} /* U_NICK */
		case U_SERVER:
		{
			char	ServName[MAXLINE];

			GetStrIndex(ServName, info, 3);
			CheckJuped(ServName);
			break;
		} /* U_SERVER */
		case U_RPLNAMES: /* yikes this is messy =/ */
		{
			char	sendstr[MAXLINE];
			char	nicks[MAXLINE];
			char	tempnix[MAXLINE];
			char	temp[MAXLINE];
			char	nick[MAXLINE];
			char	finalnix[MAXLINE]; /* nicks to deop/op */
			char	dchan[MAXLINE]; /* channel being secured */
			int		ii, cnt, userindex, dcnt, ocnt;
			
			for (ii = 0; ii < MAXLINE; ii++)
			{
				nicks[ii] = '\0';
				tempnix[ii] = '\0';
				temp[ii] = '\0';
				nick[ii] = '\0';
				finalnix[ii] = '\0';
			}
			GetStrIndex(dchan, info, 5);
			GetStrLRange(nicks, info, 6);
			strcpy(tempnix, nicks);
			GetStrRange(nicks, tempnix, 1, strlen(tempnix));
			cnt = 1;
			dcnt = 0;
			while (1)
			{
				strcpy(tempnix, nicks);
				GetStrIndex(temp, tempnix, cnt);
				if (temp[0] == '\0')
					break;
				if (temp[0] == '@')
				{
					GetStrRange(nick, temp, 1, strlen(temp));
					if (strcmp(nick, SERVICENICK) != 0)
					{
						userindex = GetUser(nick);
						if ((CheckAccess(userindex, "f") == 1) ||
								(CheckAccess(userindex, "o") == 1) ||
								(CheckAccess(userindex, "a") == 1))
						{
							if (IsRegistered(nick) != 1)
							{
								strcat(finalnix, nick);
								strcat(finalnix, " ");
								dcnt++;
							}
						}		
						else		
						{
							strcat(finalnix, nick);
							strcat(finalnix, " ");
							dcnt++;
						}	
					}	
				}
				for (ii = 0; ii < MAXLINE; ii++)
				{
					tempnix[ii] = '\0';
					temp[ii] = '\0'; 
				}	
				cnt++;
				if (dcnt == MAXMODES)
				{
					(void)sprintf(sendstr, ":%s MODE %s -oooooo %s\n",
						SERVICENICK,
						dchan,
						finalnix);
					writesocket(Servers[CurrSIndex].socket, sendstr);	
					dcnt = 0;
					for (ii = 0; ii < MAXLINE; ii++)
						finalnix[ii] = '\0';
				}
			}
			(void)sprintf(sendstr, ":%s MODE %s -oooooo %s\n",
				SERVICENICK,
				dchan,
				finalnix);
			writesocket(Servers[CurrSIndex].socket, sendstr);

			for (ii = 0; ii < MAXLINE; ii++)
			{
				nicks[ii] = '\0';
				tempnix[ii] = '\0';
				temp[ii] = '\0';
				nick[ii] = '\0';
				finalnix[ii] = '\0';
			}
			GetStrLRange(nicks, info, 6);
			strcpy(tempnix, nicks);
			GetStrRange(nicks, tempnix, 1, strlen(tempnix));
			cnt = 1;
			ocnt = 0;
			while (1)
			{
				strcpy(tempnix, nicks);
				GetStrIndex(temp, tempnix, cnt);
				if (temp[0] == '\0')
					break;
				if (temp[0] != '@')
				{
					if (temp[0] == '+')
						GetStrRange(nick, temp, 1, strlen(temp));
					else
						strcpy(nick, temp);	
					if (strcmp(nick, SERVICENICK) != 0)
					{
						userindex = GetUser(nick);
						if ((CheckAccess(userindex, "f") == 1) ||
								(CheckAccess(userindex, "o") == 1) ||
								(CheckAccess(userindex, "a") == 1))
						{
							if (IsRegistered(nick) == 1)
							{
								strcat(finalnix, nick);
								strcat(finalnix, " ");
								ocnt++;
							}
						}	
					}	
				}
				for (ii = 0; ii < MAXLINE; ii++)
				{
					tempnix[ii] = '\0';
					temp[ii] = '\0'; 
				}	
				cnt++;
				if (ocnt == MAXMODES)
				{
					(void)sprintf(sendstr, ":%s MODE %s +oooooo %s\n",
						SERVICENICK,
						dchan,
						finalnix);
					writesocket(Servers[CurrSIndex].socket, sendstr);	
					ocnt = 0;
					for (ii = 0; ii < MAXLINE; ii++)
						finalnix[ii] = '\0';
				}
			}
			(void)sprintf(sendstr, ":%s MODE %s +oooooo %s\n",
				SERVICENICK,
				dchan,
				finalnix);
			writesocket(Servers[CurrSIndex].socket, sendstr);
			break;
		} /* U_RPLNAMES */
		case U_EONAMES:
		{
			char	nchan[MAXLINE];
			char	sendstr[MAXLINE];
			int		ii;

			GetStrIndex(nchan, info, 4);
			for (ii = 0; ii < MAXCHANS; ii++)
			{
				if (channels[ii].ChanName[0] == '\0')
					continue;
				if (strcasecmp(nchan, channels[ii].ChanName) == 0)
				{
					/* it is a channel being monitored - don't part */
					return;
				}
			}
			sprintf(sendstr, ":%s PART %s\n", 
				SERVICENICK,
				nchan);
			writesocket(Servers[CurrSIndex].socket, sendstr);
			break;
		} /* U_EONAMES */
		case U_RPLMODE:
		{
			char	sendstr[MAXLINE];
			char	chanmodes[MAXLINE];
			char	donemodes[MAXLINE];
			char	mchan[MAXLINE];
			char	chankey[MAXLINE];
			int		ii, jj, waslimit, dcnt;
			int		MINUS = 0;


			for (ii = 0; ii < MAXLINE; ii++)
			{
				chanmodes[ii] = '\0';
				donemodes[ii] = '\0';
				chankey[ii] = '\0';
				mchan[ii] = '\0';
			}
			GetStrIndex(mchan, info, 4);
			GetStrLRange(chanmodes, info, 5);
			for (ii = 0; ii < MAXCHANS; ii++)
			{
				if (channels[ii].ChanName[0] == '\0')
					continue;
				if (strcasecmp(mchan, channels[ii].ChanName) == 0)
				{
					/* it is a channel being monitored, not secured so check
						 if the modes conflict with the enforced modes */
					ClearBadModes(ii, chanmodes, nick);
					
					return;
				}
			}

			/* channel is being secured, so clear all bad modes (smlik) */
			waslimit = 0;
			for (ii = 0; ii < MAXLINE; ii++)
			{
				if (chanmodes[ii] == '\0')
					break;
				if (chanmodes[ii] == 's')
					strcat(donemodes, "s");
				if (chanmodes[ii] == 'm')
					strcat(donemodes, "m");
				if (chanmodes[ii] == 'i')	
					strcat(donemodes, "i");
				if (chanmodes[ii] == 'l')
				{
					waslimit = 1;
					strcat(donemodes, "l");
				}	
				if (chanmodes[ii] == 'k')
				{
					if (waslimit == 1)
						GetStrIndex(chankey, info, 7);
					else
						GetStrIndex(chankey, info, 6);	
					strcat(donemodes, "k ");
					strcat(donemodes, chankey);
				}		
			}
			(void)sprintf(sendstr, 
				":%s MODE %s +ntp-%s\n",
				MyName,
				mchan,
				donemodes);
			writesocket(Servers[CurrSIndex].socket, sendstr);	
			break;
		} /* U_RPLMODE */
		case U_KILL: /* whenever someone /kills OBServ, bring it back */
		{
			char	killnick[MAXLINE];
			int		ii;

			for (ii = 0; ii < MAXLINE; ii++)
				killnick[ii] = '\0';
			GetStrIndex(killnick, info, 3);
			if (strcmp(killnick, SERVICENICK) == 0)
			{
				putlog(LOG1, "I was killed by %s!%s, re-initializing",
					nick,
					Lusers[luserindex].Userhost);
				InitServs();
			}
			break;
		} /* U_KILL */
		case U_KICK: /* if someone kicks OBServ, make it rejoin */
		{
			char	kchan[MAXLINE];
			char	knick[MAXLINE];
			char	kreason[MAXLINE];

			GetStrIndex(kchan, info, 3);
			GetStrIndex(knick, info, 4);
			GetStrIndex(kreason, info, 5);
			GetStrRange(kreason, kreason, 1, strlen(kreason));
			if (strncasecmp(knick, SERVICENICK, strlen(SERVICENICK)) == 0)
			{
				join(kchan);
				putlog(LOG2, "%s: %s!%s kicked me [%s]",
					kchan,
					nick,
					Lusers[luserindex].Userhost,
					kreason);
			}
			else
				putlog(LOG2, "%s: %s!%s kicked %s [%s]",
					kchan,
					nick,
					Lusers[luserindex].Userhost,
					knick,
					kreason);
			break;
		} /* U_KICK */
		case U_MODE:
		{
			int		ii;
			char	modes[MAXLINE], chan[MAXLINE];
			char	margs[MAXLINE];
			char	dnick[MAXLINE];
			char	sendstr[MAXLINE];
			int		mptr, dindex;
			int		MINUS = 0;

			GetStrIndex(chan, info, 3);
			if (!((chan[0] == '#') || (chan[0] == '&')))
				break; /* It was a umode, not a channel mode */
			GetStrLRange(modes, info, 4);
			GetStrLRange(margs, modes, 2);
			mptr = 0;
			dindex = 0;

			/* This tricky little routine *should* determine if OBServ has
				 been deoped.  If so, reop it.  *sigh* */
			for (ii = 0; ii < strlen(modes); ii++)
			{
				if (modes[mptr] == '-')
				{
					MINUS = 1;
					mptr++;
					continue;
				}
				if (modes[mptr] == '+')
				{
					MINUS = 0;
					mptr++;
					continue;
				}
				if (MINUS == 1)
				{
					if (modes[mptr] == 'o')
					{
						dindex++;
						GetStrIndex(dnick, margs, dindex);
						if (strcasecmp(dnick, SERVICENICK) == 0)
						{
							sprintf(sendstr, ":%s MODE %s +o %s\n",
								MyName,
								chan,
								SERVICENICK);
							writesocket(Servers[CurrSIndex].socket, sendstr);
							putlog(LOG1, "%s: %s!%s attempted to deop me",
								chan,
								nick,
								Lusers[luserindex].Userhost);
							break;
						}
						mptr++;
					}
					else
						mptr++;
				}
				else
				{
					if ((modes[mptr] == 'k') || (modes[mptr] == 'l') ||
							(modes[mptr] == 'v') || (modes[mptr] == 'b'))
						dindex++;
					mptr++;
				}
			}

			for (ii = 0; ii < strlen(nick); ii++)
				if (nick[ii] == '.')
					return; /* a server set the mode, allow it */

			modes[strlen(modes) - 1] = '\0';
			putlog(LOG2, "%s: mode change \"%s\" by %s!%s",
				chan,
				modes,
				nick,
				Lusers[luserindex].Userhost);

			/* check if the channel is being monitored, if so, clear any
				 conflicting modes */
			for (ii = 0; ii < MAXCHANS; ii++)
			{
				if (channels[ii].ChanName[0] == '\0')
					continue;
				if (strcasecmp(chan, channels[ii].ChanName) == 0)
				{
					ClearBadModes(ii, modes, nick);
					break;
				}
			}

			break;
		} /* U_MODE */
		case U_RPLBAN: /* clear the ban if it matches an admin */
		{
			char	tempban[MAXLINE];
			char	tempstr[MAXLINE];
			char	channel[MAXLINE];
			int		ii, jj;

			GetStrIndex(tempban, info, 5);
			GetStrIndex(channel, info, 4);
			for (ii = 0; ii < MAXUSERS; ii++)
			{
				if (Users[ii].UserHost[0] == '\0')
					continue;
				if (MatchFlags(Users[ii].UserFlags, "a") == 1)
				{
					sprintf(tempstr, "*!%s", Users[ii].UserHost);
					if (match(tempban, tempstr) == 0)
					{
						sprintf(tempstr, ":%s MODE %s -b %s\n",
							SERVICENICK,
							channel,
							tempban);
						writesocket(Servers[CurrSIndex].socket, tempstr);
						putlog(LOG1, "%s: Ban %s matchs administrator host [%s], clearing", 
							channel,
							tempban,
							Users[ii].UserHost);
						break;
					}
				}
			}
			break;
		} /* U_RPLBAN */
	}	/* switch (token) */
} /* ProcessInfo() */

void IsBadBan(char *ban, char *channel)

{
	int		ii;
	char	tempstr[MAXLINE];

	for (ii = 0; ii < MAXUSERS; ii++)
	{
		if (Users[ii].UserHost[0] == '\0')
			continue;
		if (MatchFlags(Users[ii].UserFlags, "a") == 1)
		{
			sprintf(tempstr, "*!%s", Users[ii].UserHost);
			if (match(ban, tempstr) == 0)
			{
				sprintf(tempstr, ":%s MODE %s -b %s\n",
					SERVICENICK,
					channel,
					ban);
				writesocket(Servers[CurrSIndex].socket, tempstr);
				putlog(LOG1, "%s: Ban %s matchs administrator host [%s], clearing", 
					channel,
					ban,
					Users[ii].UserHost);
				break;
			}
		}
	}
} /* IsBadBan() */

/*
IsBadMode()
	args: char *emodes, char *checkmode
	purpose: determine if 'checkmode' violates modes in 'emodes'
					 Example: if 'emodes' contains "-i" and checkmode is
					 "i", there is a violation
	return: 1 if violation found, 0 if not
*/

int IsBadMode(char *emodes, char *checkmode)

{
	int		ii;
	char	temp1[MAXLINE], temp2[MAXLINE];
	int		MINUS = 0;
	
	strcpy(temp2, checkmode);
	for (ii = 0; ii < MAXLINE; ii++)
	{
		if (emodes[ii] == '-')
		{
			MINUS = 1;
			continue;
		}
		if (emodes[ii] == '+')
		{
			MINUS = 0;
			continue;
		}
		if (MINUS == 1)
		{
			temp1[0] = emodes[ii];
			temp1[1] = '\0';
			if (strcasecmp(temp1, temp2) == 0)
				return 1;
		}
	}
	return 0;
} /* IsBadMode() */

/*
ClearBadModes()
	args: int chanindex, char *modes
	purpose: check if 'modes' conflicts with 
					 'channels[chanindex].ChanModes'; if so, clear all conflicting
					 modes
	return: none
*/

void ClearBadModes(int chanindex, char *modes, char *nick)

{
	char	donemodes[MAXLINE], chankey[MAXLINE];
	char	sendstr[MAXLINE];
	char	ban[MAXLINE];
	int		MINUS = 0, jj, dcnt, waslimit = 0;
	int		dindex, uindex;

	for (jj = 0; jj < MAXLINE; jj++)
	{
		donemodes[jj] = '\0';
		chankey[jj] = '\0';
	}

	uindex = GetUser(nick);
	if (uindex != -1)
		if ((CheckAccess(uindex, "f") == 1) ||
				(CheckAccess(uindex, "o") == 1) ||
				(CheckAccess(uindex, "a") == 1))
			return; /* allow the user to set the mode */

	donemodes[0] = '+';
	dcnt = 1;
	dindex = 2;
	for (jj = 0; jj < strlen(channels[chanindex].ChanModes); jj++)
	{
		if (channels[chanindex].ChanModes[jj] == '-')
		{
			MINUS = 1;
			continue;
		}
		if (channels[chanindex].ChanModes[jj] == '+')
		{
			MINUS = 0;
			continue;
		}
		if (MINUS == 0)
		{
			donemodes[dcnt] = channels[chanindex].ChanModes[jj];
			dcnt++;
		}
	}
	
	strcat(donemodes, "-");
	MINUS = 0;
	for (jj = 0; jj < strlen(modes); jj++)
	{
		if ((modes[jj] == '\0') ||
				(modes[jj] == '\n') ||
				(modes[jj] == '\r') ||
				(modes[jj] == ' '))
			break;
		if (modes[jj] == '-')
		{
			MINUS = 1;
			continue;
		}
		if (modes[jj] == '+')
		{
			MINUS = 0;
			continue;
		}
		if ((modes[jj] == 'n') && (IsBadMode(channels[chanindex].ChanModes, "n")))
			strcat(donemodes, "n");
		if ((modes[jj] == 't') && (IsBadMode(channels[chanindex].ChanModes, "t")))
			strcat(donemodes, "t");
		if ((modes[jj] == 's') && (IsBadMode(channels[chanindex].ChanModes, "s")))
			strcat(donemodes, "s");
		if ((modes[jj] == 'm') && (IsBadMode(channels[chanindex].ChanModes, "m")))
			strcat(donemodes, "m");
		if ((modes[jj] == 'i') && (IsBadMode(channels[chanindex].ChanModes, "i")))
			strcat(donemodes, "i");
		if (modes[jj] == 'l') 
		{
			dindex++;
			if (IsBadMode(channels[chanindex].ChanModes, "l"))
				strcat(donemodes, "l");
		}
		if ((modes[jj] == 'o') || (modes[jj] == 'v'))
			dindex++;
		if ((modes[jj] == 'b') && (MINUS == 0))
		{
			GetStrIndex(ban, modes, dindex);
			IsBadBan(ban, channels[chanindex].ChanName);
			dindex++;
		}
		if ((modes[jj] == 'k') && (IsBadMode(channels[chanindex].ChanModes, "k")))
		{
			GetStrIndex(chankey, modes, dindex);
			strcat(donemodes, "k ");
			strcat(donemodes, chankey);
			break;
		}
	}
	sprintf(sendstr, ":%s MODE %s %s\n",
		SERVICENICK,
		channels[chanindex].ChanName,
		donemodes);
	writesocket(Servers[CurrSIndex].socket, sendstr);
} /* ClearBadModes() */

/*
ConnectRemoteTcm()
	args: int index, char *line
	purpose: accept connection on TcmSocket
	return: none
*/

void ConnectRemoteTcm(int index, char *line)

{
	int		numconns, ii;
	char	sendstr[MAXLINE];
	struct sockaddr_in RemoteAddr;
	struct hostent *RemoteHost;
	int addrlen;
	int lnth;

	if (index == (-1))
	{
		numconns = 0;
		for (ii = 0; ii <= MAXDCCCONNS; ii++)
		{
			if (connections[ii].socket != -1)
				numconns++;
		}
		if (numconns > MAXDCCCONNS) /* maximum users reached */
		{
			return;
		}
		for (ii = 0; ii < MAXDCCCONNS; ii++)
			if (connections[ii].socket == -1)
				break;
	
		addrlen = sizeof(struct sockaddr);
		if((connections[ii].socket = accept(TcmSocket,
			 (struct sockaddr *)&RemoteAddr, &addrlen)) < 0 )
		{
			putlog(LOG1, "Error in remote tcm connect on accept\n");
			return;
		}
	                               
		RemoteHost = gethostbyaddr((char *)&RemoteAddr.sin_addr.s_addr,
																4,AF_INET);
	
		if(RemoteHost)
			strncpy(connections[ii].userhost, (char *)RemoteHost->h_name, MAXLINE - 1);
		else
			strncpy(connections[ii].userhost,
							inet_ntoa(RemoteAddr.sin_addr),MAXLINE - 1);

		if (GoodTCM(connections[ii].userhost) == -1)
		{
			close(connections[ii].socket);
			connections[ii].socket = (-1);
			putlog(LOG1, "Illegal connection attempt from %s", 
				connections[ii].userhost);
			connections[ii].userhost[0] = '\0';
			return;
		}
		putlog(LOG1, "Recieved connection from %s", connections[ii].userhost);

		connections[ii].isbot = 1;
		connections[ii].ispending = 1;
	} /* if (index == (-1)) */
	else
	{
		printf("line=%s\n", line);
	}

} /* ConnectRemoteTcm() */

/*
ReadServerInfo()
	args: none
	purpose: wait for information from sockets; once an entire
					 line has been sent, call ProcessInfo() to process the line
	return: none
*/

void ReadServerInfo(void)
{
	int		SelectResult, ii, jj, kk;
	struct	timeval	TimeOut;
	char		buffer[MAXLINE];
	char		c;

	for (;;)
	{
		for (ii = 0; ii < MAXLINE; ii++)
			buffer[ii] = '\0';
		FD_ZERO(&readfds);
		TimeOut.tv_sec = SERVTIMEOUT;
		TimeOut.tv_usec = 0L;
		FD_SET(Servers[CurrSIndex].socket, &readfds);
		for (ii = 0; ii < MAXDCCCONNS; ii++)
			if (connections[ii].socket != -1)
				FD_SET(connections[ii].socket, &readfds);

		FD_SET(TcmSocket, &readfds);

		if ((SelectResult = select(FD_SETSIZE, &readfds, &nullfds, &nullfds, &TimeOut)) > 0)
		{
			for (jj = 0; jj < MAXLINE; jj++)
			{

				if (FD_ISSET(TcmSocket, &readfds))
					ConnectRemoteTcm(-1, (char *) NULL);

				/* This is for info coming from the server */
				if (FD_ISSET(Servers[CurrSIndex].socket, &readfds))
				{ 
					ii = read(Servers[CurrSIndex].socket, &c, 1);
					if (ii != 0)
					{
						buffer[jj] = c;
						if ((c == '\n') || (c == '\r'))
						{
							ProcessInfo(buffer); /* process the line */
							break;
						}
					} /* if (ii != 0) */
					else
						return; /* the connection was closed */
				} /* if (FD_ISSET(Servers[CurrSIndex].socket, &readfds)) */
				else
					break;
			} /* for (jj = 0; jj < MAXLINE; jj++) */

			for (ii = 0; ii < MAXDCCCONNS; ii++)
			{
				if (connections[ii].socket == -1)
					continue;
				for (kk = 0; kk < MAXLINE; kk++)
					buffer[kk] = '\0';
				for (jj = 0; jj < MAXLINE; jj++)
				{
					/* This is for info coming from one of the DCC clients */
					if (FD_ISSET(connections[ii].socket, &readfds))
					{
						kk = read(connections[ii].socket, &c, 1);
						if (kk != 0)
						{
							buffer[jj] = c;
							if ((c == '\n') || (c == '\r'))
							{
								if (connections[ii].isbot == 1)
									BotProcess(ii, buffer); /* process line from bot */
								else
									DCCProcess(ii, buffer); /* process line from client */
								break;
							}
						} /* if (kk != 0) */
						else /* DCC client closed connection */
						{
							CloseConnection(ii);
							break;
						}
					} /* if (FD_ISSET(connections[ii].socket, &readfds)) */
				} /* for (jj = 0; jj < MAXLINE; jj++) */
			} /* for (ii = 0; ii < MAXDCCCONNS; ++ii) */
		}
		else
		{
			if (SelectResult == 0)
			{
#ifdef DEBUGMODE			
				printf("Connection closed: Ping timeout\n");
#endif
				return;
			}
		}
	} /* for (;;) */
} /* ProcessInfo() */

main(int argc, char *argv[])
{
	int			ServTCPPort; /* port to connect at */
	int			cnt, /* keeps track of what server OBServ is connected to */
					ii, /* looping */
					pid; /* pid of this process */
	char		ConfFile[MAXFILELENGTH]; /* name of config file */

	if ((argc > 1) && (strcmp(argv[1], "-h") == 0))
	{
		printf("OBServ TS services version %s by SideWnder (wnder@underworld.net)\n\n", version);
		printf("    %s -h     : Display this help screen\n", argv[0]);
		printf("    %s <file> : Specify <file> as services config file (default is observ.conf\n\n", argv[0]);
		exit(1);
	}
	printf("\nOBServ TS services version %s by SideWnder (wnder@underworld.net)\n", version);
	putlog(LOG1, "OBServ TS services started");
	ConfFile[0] = '\0'; 
	for (ii = 0; ii < MAXCHANS; ii++)
	{
		channels[ii].ChanName[0] = '\0';
		channels[ii].ChanBans[0] = '\0';
	}
	if (argc < 2)
		strcpy(ConfFile, CONFIGFILE);
	else 
		strcpy(ConfFile, argv[1]);
	strcpy(ConfigFile, ConfFile);	
	LoadConfig(ConfFile); /* load server, jupe, gline, user, admin info */

	signal(SIGHUP, ProcessSignal);
	signal(SIGSEGV, ProcessSignal);
	signal(SIGBUS, ProcessSignal);
	signal(SIGTERM, ProcessSignal);

	for (ii = 0; ii < MAXLUSERS; ii++)
		Lusers[ii].Nick[0] = '\0'; /* initialize user database */
	for (ii = 0; ii < MAXDCCCONNS; ii++)
	{
		connections[ii].socket = -1;
		connections[ii].ispending = 1;
		connections[ii].isbot = 0;
		connections[ii].ishub = 0;
	}

#ifndef DEBUGMODE
	pid = fork();
	if (pid == -1)
	{
		printf("Unable to fork(), exiting.\n");
		exit(0);
	}
	if (pid != 0)
	{
		printf("Running in background (pid: %d)\n\n", pid);
		exit(0);
	}
#else
	printf("Entering debug mode\n");	
#endif

	InitListenPort();

	cnt = 0;
	while (1)
	{
		if (Servers[cnt].ServHost[0] == '\0')
			/* we've reached the end of the server list, start at the */
			/* beginning :-) 																					*/
			cnt = 0; 

		ServTCPPort  = atoi(Servers[cnt].ConnectPort);
		if ((Servers[cnt].socket = ConnectToServer(Servers[cnt].ServHost, ServTCPPort)) > 0) /* Attempt connection to server at port 'ServTCPPort' */
		{
			CurrSIndex = cnt;
			signon();

			/* whenever OBServ connects/reconnects to a server, clear */
			/* registered users																			 	*/
			ClearReggedUsers(); 

			ReadServerInfo(); /* enter loop waiting for server info */
		}
		else 
			close(Servers[cnt].socket); /* There was an error */
		cnt++;
	}
	exit(0);
}
