/*
 * OBServ TS Services, Copyright (C) 1998 Patrick Alken
 * This program comes with absolutely NO WARRANTY
 *
 * Should you choose to use and/or modify this source code, please
 * do so under the terms of the GNU General Public License under which
 * this program is distributed.  You know the drill.
 */

#include	<stdio.h>
#include	<string.h>
#include	<sys/types.h>
#include	"config.h"
#include	"observ.h"

extern		int		CurrSIndex;
extern		char	ConfigFile[MAXFILELENGTH];
extern		char	MyName[MAXLINE];
extern		char	AdminInfo[MAXLINE];
extern		char	*version;

/*
GetStrIndex()
	args: char *str, int index
	purpose: store the 'nth' word (specified by 'index') in 'str'
	return: 'retstr', containing word 'index' of 'str'
*/

void GetStrIndex(char *retstr, char *str, int index)
{
	int		ii;
	char	*tempstr;
	char	temp[MAXLINE];
	
	strcpy(temp, str);
	tempstr = strtok(temp, " ");
	for (ii = 1; ii < index; ii++)
		tempstr = strtok((char *) NULL, " ");

	if (tempstr != NULL)
		strcpy(retstr, tempstr);
	else
		retstr[0] = '\0';
}

/*
GetStrLRange()
	args: char *donestr, char *str, int index
	purpose: similar to GetStrIndex, only returns from 'index' to EOL 
					 result is stored in 'donestr'
	return: donestr, containing the words from 'index' of 'str' to EOL				 
*/

void GetStrLRange(char *donestr, char *str, int index)
{
	int		ii;
	char	*tempstr;
	char	*temp2;
	char	temp[MAXLINE];

	strcpy(temp, str);
	tempstr = strtok(temp, " ");
	for (ii = 1; ii < index; ii++)
	{
		if (ii == index - 1)
		{
			temp2 = strtok((char *) NULL, "\r\n");
			break;
		}	
		tempstr = strtok((char *) NULL, " ");
	}
	if (temp2 != NULL)
		strcpy(donestr, temp2);
	else
		donestr[0] = '\0';	
}

/*
GetStrRange()
	args: char *str, int start, int stop
	purpose: to store a substring of 'str' ranging from index position
					 'start' to 'stop'
	return: substring described above				 
*/

void GetStrRange(char *donestr, char *str, int start, int stop)
{
	int		ii, cnt;
	char	tempstr[MAXLINE];
	char	temp[MAXLINE];

	strcpy(temp, str);
	for (ii = 0; ii < MAXLINE; ii++)
		tempstr[ii] = '\0'; 
	cnt = 0;
	for (ii = start; ii <= stop; ii++)
	{
		if (temp[ii] != '\n' && temp[ii] != '\r')
			tempstr[cnt] = temp[ii];
		else
		{
			strcpy(donestr, tempstr);
			return;
		}	
		cnt++;
	}
	strcpy(donestr, tempstr);
}

/*
StrTolower()
	args: char *retstr, char *str
	purpose: convert the first word in 'str' to lowercase and store
					 in 'retstr' 
	return: retstr, which contains 'str' only with an uppercase first 
					word
*/

void StrTolower(char *retstr, char *str)
{
	int		ii;

	strcpy(retstr, str);
	for (ii = 0; ii < strlen(str); ii++)
	{
		if (str[ii] == ' ')
			break;
		retstr[ii] = tolower(str[ii]);
	}
}

/*
AddUserInfo()
	args: char *nick, char *uhost;
	purpose: store 'nick' and 'uhost' information in next available
					 index position of structure Lusers
	return: none				 
*/

void AddUserInfo(char *nick, char *uhost) 
{
	int		ii;
	
	for (ii = 0; ii < MAXLUSERS; ii++)
	{
		if (Lusers[ii].Nick[0] == '\0')
		{
			strcpy(Lusers[ii].Nick, nick);
			strcpy(Lusers[ii].Userhost, uhost);
			break;
		}
	}
}

/*
RemUserInfo()
	args: char *nick
	purpose: when a user either QUITs the network, or changes their NICK,
					 delete the old user from the database.
	return: none				 
*/

void RemUserInfo(char *nick)
{
	int		ii;

	for (ii = 0; ii < MAXLUSERS; ii++)
	{
		if (strcasecmp(Lusers[ii].Nick, nick) == 0)
		{
			Lusers[ii].Nick[0] = '\0';
			Lusers[ii].Userhost[0] = '\0';
			Lusers[ii].IsRegged = 0;
			break;
		}
	}
}

/*
CheckChans()
	args: struct ChanInfo oldchans[]
	purpose: this function is called after a rehash, to check if any
					 new channels have been added to the config file, or any
					 old channels removed from the config file.  Any new
					 channels are join()'d and old channels no longer
					 being monitored are part()'d.
	return: none
*/

void CheckChans(struct ChanInfo oldchans[MAXCHANS])

{
	int		ii, jj, found;

	for (ii = 0; ii < MAXCHANS; ii++)
	{
		if (oldchans[ii].ChanName[0] == '\0')
			break;
		if (IsChannel(oldchans[ii].ChanName) == -1)
		{
			part(oldchans[ii].ChanName);
			putlog(LOG1, "No longer monitoring %s", oldchans[ii].ChanName);
			oldchans[ii].ChanName[0] = '\0';
		}
	}
	for (ii = 0; ii < MAXCHANS; ii++)
	{
		if (channels[ii].ChanName[0] == '\0')
			continue;
		found = 0;
		for (jj = 0; jj < MAXCHANS; jj++)
		{
			if (oldchans[jj].ChanName[0] == '\0')
				continue;
			if (!strcasecmp(channels[ii].ChanName, oldchans[jj].ChanName))
			{
				found = 1;
				break;
			}
		}
		if (found == 0)
		{
			join(channels[ii].ChanName);
			putlog(LOG1, "Now monitoring %s [%s]", 
				channels[ii].ChanName,
				channels[ii].ChanModes);
		}
	}
}

/*
notice()
	args: char *nick, char *msg, int sockfd
	purpose: send a NOTICE to 'nick' with 'msg'
	return: none
*/

void notice(char *nick, char *msg, int sockfd)

{
	char	sendstr[MAXLINE];

	if (sockfd == NODCC)
	{
		sprintf(sendstr, ":%s NOTICE %s :%s\n", 
			SERVICENICK,
			nick,
			msg);
		writesocket(Servers[CurrSIndex].socket, sendstr);
	}
	else
	{
		sprintf(sendstr, "%s\n", msg);
		writesocket(sockfd, sendstr);
	}
} /* notice() */

/*
ServNotice()
	args: char *nick, char *msg, int sockfd
	purpose: same as notice(), only use the server name instead of
					 SERVICENICK to send the notice
	return: none
*/

void ServNotice(char *nick, char *msg, int sockfd)

{
	char	sendstr[MAXLINE];

	if (sockfd == NODCC)
	{
		sprintf(sendstr, ":%s NOTICE %s :*** %s\n", 
			MyName,
			nick,
			msg);
		writesocket(Servers[CurrSIndex].socket, sendstr);
	}
	else
	{
		sprintf(sendstr, "%s\n", msg);
		writesocket(sockfd, sendstr);
	}
} /* ServNotice() */

/*
SendToOps()
	args: char *msg
	purpose: send 'msg' to all currently registered +o or +a users
	return: none
*/

void SendToOps(char *msg)

{
#ifdef NOTIFYOPS
	int		userindex;
	int		ii;

	for (ii = 0; ii < MAXLUSERS; ii++)
	{
		if (Lusers[ii].Nick[0] == '\0')
			continue;
		if (Lusers[ii].IsRegged == YES)
		{
			userindex = GetUser(Lusers[ii].Nick);
			if ((CheckAccess(userindex, "o") == 1) ||
					(CheckAccess(userindex, "a") == 1))
			{
				ServNotice(Lusers[ii].Nick, msg, NODCC);
			}
		}
	}
#endif
}

/*
SendMOTD
	args: int sockfd
	purpose: output message of the day to the socket 'sockfd'
	return: none
*/

void SendMOTD(int sockfd)

{
	char	line[MAXLINE];
	char	sendstr[MAXLINE];
	FILE	*fp;

	if ((fp = fopen(MOTDFILE, "r")) == (FILE *) NULL)
	{
		writesocket(sockfd, "MOTD file missing\n");
		return;
	}

	sprintf(sendstr, "Welcome to OBServ %s\n", version);
	writesocket(sockfd, sendstr);

	while (fgets(line, MAXLINE - 1, fp))
		writesocket(sockfd, line);

	fclose(fp);
} /* SendMOTD() */

/*
Substitute()
	args: char *str
	purpose: replace any formatting characters in 'str' with the
					 corresponding information.  Stores newly formatted string
					 back into 'str'.

		Formatting characters:
			%S - Nickname of services
			%A - Administrative info
			%V - current version
			%B - bold character

	return: none
*/

void Substitute(char *str)

{
	char	tempstr[MAXLINE], key;
	char	finalstr[MAXLINE];
	char	*ptr;
	int		tcnt, fcnt;

	strcpy(tempstr, str);
	for (fcnt = 0; fcnt < MAXLINE; fcnt++)
		finalstr[fcnt] = '\0';
	fcnt = 0;
	tcnt = 0;
	while (tcnt < MAXLINE)
	{
		if ((tempstr[tcnt] == '\n') || (tempstr[tcnt] == '\r'))
			break;
		if (tempstr[tcnt] == '%')
		{
			key = tempstr[++tcnt];
			switch (key)
			{
				case 's':
				case 'S':
				{
					strcat(finalstr, SERVICENICK);
					fcnt += strlen(SERVICENICK) - 1;
					break;
				}
				case 'b':
				case 'B':
				{
					strcat(finalstr, "\002");
					break;
				}
				case 'v':
				case 'V':
				{
					strcat(finalstr, version);
					fcnt += strlen(version) - 1;
					break;
				}
				case 'a':
				case 'A':
				{
					strcat(finalstr, AdminInfo);
					fcnt += strlen(AdminInfo) - 1;
					break;
				}

				default:
				{
					strcat(finalstr, "%");
					finalstr[++fcnt] = tempstr[tcnt];
					break;
				}
			} /* switch (key) */
		}
		else
			finalstr[fcnt] = tempstr[tcnt];
		fcnt++;
		tcnt++;
	}
	strcpy(str, finalstr);
} /* Substitute() */

/*
GiveHelp()
	args: char *helpnick, char *command
	purpose: give 'helpnick' help on the command specified in 'command';
					 if no command is given, output a list of available commands
	return: none				 
*/

void GiveHelp(char *helpnick, char *command, int sockfd)
{
	FILE	*fp;
	char	sendstr[MAXBUFF];
	char	line[MAXLINE];
	char	helparg[MAXLINE];
	int		token;

	StrTolower(helparg, command);
	if (helparg[0] == '\0') /* they want general help */
	{
		if (sockfd == NODCC) /* they want /msg commands */
		{
			if ((fp = fopen(MSGHELPFILE, "r")) == NULL)
			{
				ServNotice(helpnick, "Unable to open help file", sockfd);
				return;
			}
		}
		else /* they want dcc commands */
		{
			if ((fp = fopen(DCCHELPFILE, "r")) == NULL)
			{
				ServNotice(helpnick, "Unable to open help file", sockfd);
				return;
			}
		}

		while (fgets(line, MAXLINE - 1, fp))
		{
			if (line[0] == '\n')
			{
				if (sockfd != NODCC)
					notice(helpnick, "\r\n", sockfd);
				else
					notice(helpnick, "\002", NODCC);
				continue;
			}
			if ((line[0] == '\r') || (line[0] == '\n'))
				continue;
			Substitute(line);
			notice(helpnick, line, sockfd);
		}
		fclose(fp);
	}
	else /* they want help on a specific command */
	{
		if (sockfd == NODCC)
			sprintf(sendstr, "%smsg/%s", HELPPATH, helparg);
		else
			sprintf(sendstr, "%sdcc/%s", HELPPATH, helparg);
		if ((fp = fopen(sendstr, "r")) == (FILE *)NULL)
		{
			sprintf(sendstr, "No help available on %s", command);
			notice(helpnick, sendstr, sockfd);
			return;
		}
		while (fgets(line, MAXLINE - 1, fp))
		{
			Substitute(line);
			notice(helpnick, line, sockfd);
		}
		fclose(fp);
	}	/* else */
} /* GiveHelp() */

/*
ClearReggedUsers()
	args: none
	purpose: unregister all users who have access to OBServ
	return: none
*/

void ClearReggedUsers()
{
	int		ii;

	for (ii = 0; ii < MAXLUSERS; ii++)
		Lusers[ii].IsRegged = NO;
}

/*
IsRegistered()
	args: char *nick;
	purpose: check if 'nick' has registered with OBServ
	return: 1 if nick has registered, -1 if not
*/

int IsRegistered(char *nick)
{
	int		ii;

	for (ii = 0; ii < MAXLUSERS; ii++)
	{
		if (strcmp(nick, Lusers[ii].Nick) == 0)
			if (Lusers[ii].IsRegged == YES)
				return(YES);
	}
	return(-1);
}

/*
ProcessCommand()
	args: char nick[MAXLINE], char command[MAXLINE]
	purpose: process command 'command' sent by 'nick'
	return: none
*/

void ProcessCommand(char *nick, char *command, int sockfd)

{
	int		ii;
	char	keyword[MAXLINE];
	int		token = 0, luserindex;
	char	userhost[MAXLINE];
	int		userindex;

	GetStrIndex(keyword, command, 1);

	/* Just assume IsUser() will not return -1, since the only way this */
	/* function will be called is if the nick is already registered,  	*/
	/* or matches a U: line in the config file													*/
	luserindex = IsLuser(nick);
	userindex = GetUser(nick);
	strcpy(userhost, Users[userindex].UserHost);	
	if (strncasecmp(keyword, "REGISTER", 8) == 0)
		token = C_REGISTER;
	if (strncasecmp(keyword, "KILL", 4) == 0)
		token = C_KILL;
	if (strncasecmp(keyword, "REHASH", 6) == 0)
		token = C_REHASH;
	if (strncasecmp(keyword, "JUPE", 4) == 0)
		token = C_JUPE;	
	if (strncasecmp(keyword, "OMODE", 5) == 0)
		token = C_OMODE;
	if (strncasecmp(keyword, "SECURE", 6) == 0)
		token = C_SECURE;
	if (strncasecmp(keyword, "DIE", 3) == 0)
		token = C_DIE;
	if (strncasecmp(keyword, "STATUS", 6) == 0)
		token = C_STATUS;
	if (strncasecmp(keyword, "UNJUPE", 6) == 0)
		token = C_UNJUPE;
	if (strncasecmp(keyword, "GLINE", 5) == 0)
		token = C_GLINE;
	if (strncasecmp(keyword, "UNGLINE", 7) == 0)
		token = C_UNGLINE;
	if (strncasecmp(keyword, "HELP", 4) == 0) 
		token = C_HELP; /* lets give this user some help :-) */
	if (strncasecmp(keyword, "JOIN", 4) == 0)
		token = C_JOIN;
	if (strncasecmp(keyword, "PART", 4) == 0)
		token = C_PART;
	if (strncasecmp(keyword, "OP", 2) == 0)
		token = C_OP;

	switch (token)
	{
		case C_REGISTER:
		{
			char	sendstr[MAXBUFF];
			char	pass[MAXLINE];

			if (strcmp(Users[userindex].UserHost, userhost) == 0)
			{
				GetStrIndex(pass, command, 2);
				if (pass[0] == '\0')
				{
					notice(nick, "Usage: REGISTER <password>", sockfd);
					return;	
				}
				if (IsRegistered(nick) == 1)
				{
					notice(nick, "You are already registered", sockfd);
					return;
				}
				if (strcmp(Users[userindex].UserPass, pass) == 0)
				{
					notice(nick, "You are now registered", sockfd);
					Lusers[luserindex].IsRegged = YES;
					if (sockfd == NODCC)
						sprintf(sendstr, "(%s!%s) !*! REGISTER ...\n",
							Lusers[luserindex].Nick,
							Lusers[luserindex].Userhost);
					else
						sprintf(sendstr, "#%s# register ...\n",
							Lusers[luserindex].Nick);
					BroadcastDCC(-1, sendstr, DCCOPS);
					if (sockfd != NODCC)
					{
						sprintf(sendstr, "%s@%s has registered\n", 
							nick,
							SERVICENICK);
						BroadcastDCC(-1, sendstr, DCCOPS);
					}
				}
				else 
				{
					notice(nick, "Access Denied", sockfd);
					if (sockfd == NODCC)
						sprintf(sendstr, "(%s!%s) failed REGISTER\n",
							Lusers[luserindex].Nick,
							Lusers[luserindex].Userhost);
					else
						sprintf(sendstr, "#%s# failed REGISTER\n",
							Lusers[luserindex].Nick);
					BroadcastDCC(-1, sendstr, DCCOPS);
					putlog(LOG1, "[%s!%s] failed REGISTER", 
						Lusers[luserindex].Nick,
						Lusers[luserindex].Userhost);
					return;
				}
			}
			break;
		} /* C_REGISTER */
		case C_KILL:
		{
			char	sendstr[MAXBUFF];
			char	knick[MAXLINE];
			char	kreason[MAXLINE], origreason[MAXLINE];
			int		kindex, klindex;

			if ((CheckAccess(userindex, "o") == 1) ||
					(CheckAccess(userindex, "a") == 1))
			{
				GetStrIndex(knick, command, 2);
				GetStrLRange(kreason, command, 3);
				kindex = GetUser(knick);
				klindex = IsLuser(knick);
				if (CheckAccess(kindex, "e") == 1)
				{
					/* The user is protected */
					sprintf(sendstr, "%s!%s is protected", 
						knick,
						Lusers[klindex].Userhost);
					notice(nick, sendstr, sockfd);
					putlog(LOG1, "%s!%s attempted to KILL protected user %s [%s]",
						Lusers[luserindex].Nick,
						Lusers[luserindex].Userhost,
						knick,
						kreason);
					return;
				}
				if (strcasecmp(knick, SERVICENICK) == 0)
				{
					notice(nick, "I don't think so", sockfd);
					return;
				}
				strcpy(origreason, kreason);
				if (knick[0] == '\0')
				{
					notice(nick, "Usage: KILL <nick> <reason>", sockfd);
					return;
				}
				else
				{
					if (kreason[0] == '\0')
					{
						(void)sprintf(sendstr, "Requested by %s",
							Lusers[luserindex].Nick); 
						strcpy(kreason, sendstr);
					}
					else
					{
						(void)sprintf(sendstr, " (Requested by %s)",
							Lusers[luserindex].Nick);
						strcat(kreason, sendstr);	
					}
					(void)sprintf(sendstr, ":%s KILL %s :%s!%s (%s)\n",
						SERVICENICK,
						knick,
						MyName,
						SERVICENICK,
						kreason);
					writesocket(Servers[CurrSIndex].socket, sendstr);
					putlog(LOG1, "Recieved KILL request from %s!%s for %s [%s]",
						Lusers[luserindex].Nick,
						Lusers[luserindex].Userhost,
						knick,
						origreason);
					RemUserInfo(knick);	
					sprintf(sendstr, "%s has been killed [%s]\n", knick, origreason);
					notice(nick, sendstr, sockfd);
					if (sockfd == NODCC)
						sprintf(sendstr, "(%s!%s) !*! KILL %s %s\n",
							Lusers[luserindex].Nick,
							Lusers[luserindex].Userhost,
							knick,
							origreason);
					else
						sprintf(sendstr, "#%s# kill %s %s\n",
							Lusers[luserindex].Nick,
							knick,
							origreason);
					BroadcastDCC(-1, sendstr, DCCOPS);
				}		 
			}
			else
				notice(nick, "No privileges", sockfd);
			break;	
		} /* C_KILL */
		case C_REHASH:
		{
			char	sendstr[MAXLINE];
			char	ConfFile[MAXFILELENGTH];

			if (CheckAccess(userindex, "a") == 1)
			{
				strcpy(ConfFile, ConfigFile);
				if (Rehash() == 0)
				{
					notice(nick, "Rehash complete", sockfd);
					putlog(LOG1, "Rehash requested by %s [%s] (successful)",
						Lusers[luserindex].Nick,
						Lusers[luserindex].Userhost);
				}
				else
				{
					notice(nick, "Rehash failed", sockfd);
					putlog(LOG1, "Rehash requested by %s [%s] (failed)",
						Lusers[luserindex].Nick,
						Lusers[luserindex].Userhost);
				}
				if (sockfd == NODCC)
					sprintf(sendstr, "(%s!%s) !*! REHASH\n",
						Lusers[luserindex].Nick,
						Lusers[luserindex].Userhost);
				else
					sprintf(sendstr, "#%s# rehash\n", Lusers[luserindex].Nick);
				BroadcastDCC(-1, sendstr, DCCOPS);
			}
			else
				notice(nick, "No privileges", sockfd);
			break;	
		} /* C_REHASH */
		case C_JUPE:
		{
			char	JupeName[MAXLINE];
			char	JupeReason[MAXLINE];
			char	sendstr[MAXBUFF];
			char	ConfFile[MAXFILELENGTH];
			FILE	*fp;
			int		ii;
			char	dummy[MAXLINE];
			char	serv1[MAXLINE], 
						serv2[MAXLINE]; /* to check if jupe already exists */

			for (ii = 0; ii < MAXLINE; ii++)
			{
				JupeName[ii] = '\0';
				JupeReason[ii] = '\0';
				serv1[ii] = '\0';
				serv2[ii] = '\0';
				dummy[ii] = '\0';
			}	
			if ((CheckAccess(userindex, "j") == 1) ||
					(CheckAccess(userindex, "o") == 1) ||
					(CheckAccess(userindex, "a") == 1))
			{
				GetStrIndex(JupeName, command, 2);
				GetStrLRange(JupeReason, command, 3);
				if ((JupeName[0] == '\0') || (JupeReason[0] == '\0'))
				{
					notice(nick, "Usage: JUPE <server> <reason>", sockfd);
					return;	
				}
				/* Check if JUPE already exists */
				for (ii = 0; ii < MAXJUPES; ii++)
				{
					if (Jupes[ii].ServName[0] == '\0')
						continue;
					if (match(JupeName, Jupes[ii].ServName) == 0)
					{
						sprintf(sendstr, "A jupe already exists for %s [%s]",
							Jupes[ii].ServName,
							Jupes[ii].reason);
						notice(nick, sendstr, sockfd);
						return;	
					}	
				}
				strcpy(ConfFile, ConfigFile);
				if ((fp = fopen(ConfFile, "a")) == (FILE *) NULL)
				{
					notice(nick, "Unable to open config file", sockfd);
					return;
				}
				(void)sprintf(sendstr, "J:%s:%s\n", JupeName, JupeReason);
				fputs(sendstr, fp);
				fclose(fp);
				strcpy(ConfFile, ConfigFile);
				for (ii = 0; ii < MAXJUPES; ii++)
					if (Jupes[ii].ServName[0] == '\0')
					{
						strcpy(Jupes[ii].ServName, JupeName);
						strcpy(Jupes[ii].reason, JupeReason);
						break;
					}
				(void)sprintf(sendstr, "Jupe for %s [%s] has been successfully added to config file", JupeName, JupeReason);
				notice(nick, sendstr, sockfd);
				if (sockfd == NODCC)
					sprintf(sendstr, "(%s!%s) !*! JUPE %s %s\n",
						Lusers[luserindex].Nick,
						Lusers[luserindex].Userhost,
						JupeName,
						JupeReason);
				else
					sprintf(sendstr, "#%s# jupe %s %s\n",
						Lusers[luserindex].Nick,
						JupeName,
						JupeReason);
				BroadcastDCC(-1, sendstr, DCCOPS);
				putlog(LOG1, "Jupe added for %s [%s] by %s!%s",
					JupeName,
					JupeReason,
					Lusers[luserindex].Nick,
					Lusers[luserindex].Userhost);
				sprintf(sendstr, "Server jupe added by %s for %s [%s]",
					nick,
					JupeName,
					JupeReason);
				SendToOps(sendstr);
				(void)sprintf(sendstr, "SQUIT %s :Juped: %s\n",
					JupeName, JupeReason);
				writesocket(Servers[CurrSIndex].socket, sendstr);
			}
			else
				notice(nick, "No privileges", sockfd);
			break;	
		} /* C_JUPE */
		case C_OMODE:
		{
			char	sendstr[MAXBUFF];
			char	ochannel[MAXLINE];
			char	omodes[MAXLINE];
			int		ii;

			if (CheckAccess(userindex, "a") == 1)
			{
				for (ii = 0; ii < MAXLINE; ii++)
				{
					ochannel[ii] = '\0';
					omodes[ii] = '\0';
				}
				GetStrIndex(ochannel, command, 2);
				GetStrLRange(omodes, command, 3);
				if ((ochannel[0] == '\0') || (omodes[0] == '\0')) 
				{
					notice(nick, "Usage: OMODE <#channel> <modes>", sockfd);
					return;
				}
				(void)sprintf(sendstr, ":%s MODE %s %s\n",
					MyName,
					ochannel,
					omodes);
				writesocket(Servers[CurrSIndex].socket, sendstr);
				if (sockfd == NODCC)
					sprintf(sendstr, "(%s!%s) !*! OMODE %s %s\n",
						Lusers[luserindex].Nick,
						Lusers[luserindex].Userhost,
						ochannel,
						omodes);
				else
					sprintf(sendstr, "#%s# omode %s %s\n",
						Lusers[luserindex].Nick,
						ochannel,
						omodes);
				BroadcastDCC(-1, sendstr, DCCOPS);
				putlog(LOG1, "%s [%s] used OMODE (%s) on %s",
					Lusers[luserindex].Nick,
					Lusers[luserindex].Userhost,
					omodes,
					ochannel);
				sprintf(sendstr, "Administrator %s used OMODE on %s [%s]\n",
					nick,
					ochannel,
					omodes);
				SendToOps(sendstr);
			}
			else
				notice(nick, "No privileges", sockfd);
			break;
		} /* C_OMODE */
		case C_SECURE:
		{
			char	sendstr[MAXBUFF];
			char	schan[MAXLINE];
			int		ii;

			if (CheckAccess(userindex, "a") == 1)
			{
				for (ii = 0; ii < MAXLINE; ii++)
					schan[ii] = '\0';
				GetStrIndex(schan, command, 2);
				if (schan[0] == '\0')
				{
					notice(nick, "Usage: SECURE <#channel>", sockfd);
					return;
				}
				join(schan);
				sprintf(sendstr, ":%s NAMES %s\n",
					SERVICENICK,
					schan);
				writesocket(Servers[CurrSIndex].socket, sendstr);	
				if (sockfd == NODCC)
					sprintf(sendstr, "(%s!%s) !*! SECURE %s\n",
						Lusers[luserindex].Nick,
						Lusers[luserindex].Userhost,
						schan);
				else
					sprintf(sendstr, "#%s# secure %s\n",
						Lusers[luserindex].Nick,
						schan);
				BroadcastDCC(-1, sendstr, DCCOPS);
				putlog(LOG1, "%s [%s] used SECURE %s", 
					Lusers[luserindex].Nick,
					Lusers[luserindex].Userhost,
					schan);
				sprintf(sendstr, "Administrator %s has secured channel %s\n",
					nick,
					schan);
				SendToOps(sendstr);
			}
			else
				notice(nick, "No privileges", sockfd);
			break;
		} /* C_SECURE */
		case C_DIE:
		{
			char	sendstr[MAXLINE];
			char 	reason[MAXLINE];
			int		ii;

			if (CheckAccess(userindex, "a") == 1)
			{
				for (ii = 0; ii < MAXLINE; ii++)
					reason[ii] = '\0';
				GetStrLRange(reason, command, 2);
				if (reason[0] == '\0')
				{
					(void)sprintf(sendstr, "Shutdown authorized by %s", nick);
					strcpy(reason, sendstr);
				}
				else
				{
					sprintf(sendstr, " (authorized by %s)", nick);
				  strcat(reason, sendstr);
				}
				if (sockfd == NODCC)
					sprintf(sendstr, "(%s!%s) !*! DIE %s\n",
						Lusers[luserindex].Nick,
						Lusers[luserindex].Userhost,
						reason);
				else
					sprintf(sendstr, "#%s# die %s\n",
						Lusers[luserindex].Nick,
						reason);
				BroadcastDCC(-1, sendstr, DCCOPS);
				(void)sprintf(sendstr, ":%s QUIT :%s\nSQUIT %s :%s\n",
					SERVICENICK,
					"OBServ shutting down...",
					Servers[CurrSIndex].RealName,
					reason);
				writesocket(Servers[CurrSIndex].socket, sendstr);
				putlog(LOG1, "Shutting down by request of %s!%s",
					Lusers[luserindex].Nick,
					Lusers[luserindex].Userhost);
				exit(1);
			}
			else
				notice(nick, "No privileges", sockfd);
			break;
		} /* C_DIE */
		case C_STATUS:
		{
			char	sendstr[MAXLINE];
			int		ii, gcnt, jcnt, ccnt, uindex, ucnt, cnt;

			if ((CheckAccess(userindex, "a") == 1) ||
					(CheckAccess(userindex, "o") == 1))
			{
				jcnt = 0;
				for (ii = 0; ii < MAXJUPES; ii++)
				{
					if (Jupes[ii].ServName[0] == '\0')
						break;
					jcnt++;
				}
				gcnt = 0;
				for (ii = 0; ii < MAXGLINES; ii++)
				{
					if (Glines[gcnt].GHost[0] == '\0')
						break;
					gcnt++;	
				}
				ccnt = 0;
				for (ii = 0; ii < MAXCHANS; ii++)
				{
					if (channels[ii].ChanName[0] == '\0')
						break;
					ccnt++;
				}
				
				if (jcnt != 0)
				{
					(void)sprintf(sendstr, "Active Jupes: (%d)", jcnt);
					notice(nick, sendstr, sockfd);
					for (ii = 1; ii <= jcnt; ii++)
					{
						sprintf(sendstr, "  %d) %s [%s]", 
							ii, 
							Jupes[ii - 1].ServName,
							Jupes[ii - 1].reason);
						notice(nick, sendstr, sockfd);
					}
				} /* if (jcnt != 0) */
				else
					notice(nick, "No Active Jupes", sockfd);
				if (gcnt != 0)
				{
					sprintf(sendstr, "Active Glines: (%d)", gcnt);
					notice(nick, sendstr, sockfd);
					for (ii = 1; ii <= gcnt; ii++)
					{
						sprintf(sendstr, "  %d) %s [%s]",
							ii,
							Glines[ii - 1].GHost,
							Glines[ii - 1].GReason);
						notice(nick, sendstr, sockfd);
					}
				} /* if (gcnt != 0) */
				else
					notice(nick, "No Active Glines", sockfd);

			#ifdef CHANNELSERVICES
				if (ccnt != 0)
				{
					sprintf(sendstr, "Current channels being monitored: (%d)", ccnt);
					notice(nick, sendstr, sockfd);
					cnt = 1;
					for (ii = 0; ii < MAXCHANS; ii++)
					{
						if (channels[ii].ChanName[0] == '\0')
							continue;
						sprintf(sendstr, "  %d) %s [%s]", 
							cnt, 
							channels[ii].ChanName,
							channels[ii].ChanModes);
						notice(nick, sendstr, sockfd);
						cnt++;
					}
				}
				else
					notice(nick, "Not monitoring any channels", sockfd);
			#else
				notice(nick, "Channel services not enabled", sockfd);
			#endif

				notice(nick, "Current Registered Users:", sockfd);
				ucnt = 1;
				for (ii = 0; ii < MAXLUSERS; ii++)
				{
					if (Lusers[ii].Nick[0] == '\0')
						break;
					if (IsRegistered(Lusers[ii].Nick) == 1)
					{
						uindex = GetUser(Lusers[ii].Nick);
						sprintf(sendstr, "  %d) %s [%s] [%s]",
							ucnt,
							Lusers[ii].Nick,
							Lusers[ii].Userhost,
							Users[uindex].UserFlags);
						notice(nick, sendstr, sockfd);
						ucnt++;
					}
				}
				if (sockfd == NODCC)
					sprintf(sendstr, "(%s!%s) !*! STATUS\n",
						Lusers[luserindex].Nick,
						Lusers[luserindex].Userhost);
				else
					sprintf(sendstr, "#%s# status\n",
						Lusers[luserindex].Nick);
				BroadcastDCC(-1, sendstr, DCCOPS);
			}
			else
				notice(nick, "No privileges", sockfd);
			break;
		} /* C_STATUS */
		case C_UNJUPE:
		{
			char	sendstr[MAXBUFF];
			int		ii, jj, jcnt;
			char	jname[MAXLINE], jtemp[MAXLINE];
			char	line[MAXLINE], linetemp[MAXLINE];
			char	ConfName[MAXFILELENGTH];
			char	TempFName[MAXLINE] = "t!emp667.tmp";
			char	*key;
			char	*jhost;
			FILE	*configfp;
			FILE	*fp; /* random file to store all lines except the ones */
								 /* which match the jupes 												 */

			if ((CheckAccess(userindex, "j") == 1) ||
					(CheckAccess(userindex, "o") == 1) ||
					(CheckAccess(userindex, "a") == 1))
			{
				for (ii = 0; ii < MAXLINE; ii++)
					jname[ii] = '\0';
				GetStrIndex(jname, command, 2);
				if (jname[0] == '\0')
				{
					notice(nick, "Usage: UNJUPE <servername>", sockfd);
					return;
				}
				strcpy(ConfName, ConfigFile);
				if ((configfp = fopen(ConfName, "r")) == NULL)
				{
					notice(nick, "Unable to open config file", sockfd);
					return;
				}
				fp = fopen(TempFName, "w"); 
				jcnt = 0;
				while (fgets(line, MAXLINE - 1, configfp))
				{
					strcpy(linetemp, line);
					key = strtok(line, ":");
					if ((strcmp(key, "j") == 0) || (strcmp(key, "J") == 0))
					{
						jhost = strtok((char *) NULL, ":"); 
						if (match(jname, jhost) != 0)
							fputs(linetemp, fp);
						else
							jcnt++;
					}
					else
						fputs(linetemp, fp);
				} /* while () */
				fclose(configfp);
				fclose(fp); 
				(void)sprintf(sendstr, "cp %s %s", TempFName, ConfName);
				system(sendstr);
				sprintf(sendstr, "rm -f %s", TempFName);
				system(sendstr);
				for (ii = 0; ii < MAXJUPES; ii++)
				{
					if (Jupes[ii].ServName[0] == '\0')
						break;
					if (wldwld(jtemp, jname) == 0)
						for (jj = 0; jj < MAXLINE; jj++)
							Jupes[ii].ServName[jj] = '\0';
				}
				GetStrIndex(jname, command, 2);
				if (jcnt == 0)
				{
					sprintf(sendstr, "No jupes matching %s found", jname);
					notice(nick, sendstr, sockfd);
					return;			
				} 
				for (ii = 0; ii < MAXJUPES; ii++)
				{
					if (Jupes[ii].ServName[0] == '\0')
						continue;
					if (wldwld(jname, Jupes[ii].ServName) == 0)
						Jupes[ii].ServName[0] = '\0';
				}
				sprintf(sendstr, "Jupes for %s removed [%d matches]\n", jname, jcnt);
				notice(nick, sendstr, sockfd);
				if (sockfd == NODCC)
					sprintf(sendstr, "(%s!%s) !*! UNJUPE %s\n",
						Lusers[luserindex].Nick,
						Lusers[luserindex].Userhost,
						jname);
				else
					sprintf(sendstr, "#%s# unjupe %s\n",
						Lusers[luserindex].Nick,
						jname);
				BroadcastDCC(-1, sendstr, DCCOPS);
				putlog(LOG1, "Server jupes matching %s removed by %s!%s",
					jname,
					Lusers[luserindex].Nick,
					Lusers[luserindex].Userhost);
				sprintf(sendstr, "Server jupes matching %s removed by %s",
					jname,
					nick);
				SendToOps(sendstr);
			}
			else
				notice(nick, "No privileges", sockfd);
			break;
		} /* C_UNJUPE */
		case C_GLINE:
		{
			char	sendstr[MAXBUFF];
			char	ghost[MAXLINE];
			char	greason[MAXLINE];
			int		ii, ufound, dfound;
			FILE	*fp;
			char	ConfFile[MAXFILELENGTH];

			if (CheckAccess(userindex, "g") == 1)
			{
				for (ii = 0; ii < MAXLINE; ii++);
				{
					ghost[ii] = '\0';
					greason[ii] = '\0';
				}
				GetStrIndex(ghost, command, 2);
				GetStrLRange(greason, command, 3);
				if (greason[0] == '\0')
				{
					notice(nick, "Usage: GLINE <user@host> <reason>", sockfd);
					return;
				}
				if (strcmp(ghost, "*@*") == 0)
				{
					notice(nick, "Cannot gline *@*", sockfd);
					return;	
				}

				/* Check if a valid user@host was given */
				ufound = 0;
				dfound = 0;
				for (ii = strlen(ghost); ii > 0; ii--)
				{
					if (ghost[ii] == '.')
						dfound = 1;
					if (ghost[ii] == '@')
						ufound = 1;	
				}
				if (!ufound) /* there was no username specified */
				{
					sprintf(sendstr, "Please include a username with the host.  (i.e. username@%s)", ghost);
					notice(nick, sendstr, sockfd);
					return;
				}
				if (!dfound)
				/* The hostname has no '.' in it, consider it invalid */
				{ 
					sprintf(sendstr, "%s does not contain a valid hostname", ghost);
					notice(nick, sendstr, sockfd);
					return;	
				}
					
				/* Check if ghost matches an exception user */
				for (ii = 0; ii < MAXUSERS; ii++)
				{
					if (Users[ii].UserHost[0] == '\0')
						break;
					if (match(ghost, Users[ii].UserHost) == 0)
						if ((CheckAccess(ii, "e") == 1)) 
						{
							sprintf(sendstr, "%s matches protected host [%s], unable to gline", ghost, Users[ii].UserHost);
							notice(nick, sendstr, sockfd);
							return;
						}	
				}

				/* Check if GLINE already exists */
				for (ii = 0; ii < MAXGLINES; ii++)
				{
					if (Glines[ii].GHost[0] == '\0')
						continue;
					if (match(ghost, Glines[ii].GHost) == 0)
					{
						(void)sprintf(sendstr, "A gline already exists for %s [%s]", Glines[ii].GHost, Glines[ii].GReason);
						notice(nick, sendstr, sockfd);
						return;	
					}	
				}
				strcpy(ConfFile, ConfigFile);
				if ((fp = fopen(ConfFile, "a")) == (FILE *) NULL)
				{
					notice(nick, "Unable to open config file", sockfd);
					return;
				}
				sprintf(sendstr, "G:%s:%s\n", ghost, greason);
				fputs(sendstr, fp);
				fclose(fp);
				strcpy(ConfFile, ConfigFile);
				for (ii = 0; ii < MAXGLINES; ii++)
				{
					if (Glines[ii].GHost[0] == '\0')
					{
						strcpy(Glines[ii].GHost, ghost);
						strcpy(Glines[ii].GReason, greason);
						break;
					}
				}
				sprintf(sendstr, "Gline for %s [%s] has been successfully added to config file", ghost, greason);
				notice(nick, sendstr, sockfd);
				if (sockfd == NODCC)
					sprintf(sendstr, "(%s!%s) !*! GLINE %s %s\n",
						Lusers[luserindex].Nick,
						Lusers[luserindex].Userhost,
						ghost,
						greason);
				else
					sprintf(sendstr, "#%s# gline %s %s\n",
						Lusers[luserindex].Nick,
						ghost,
						greason);
				BroadcastDCC(-1, sendstr, DCCOPS);
				putlog(LOG1, "Gline for %s [%s] added by %s!%s",
					ghost,
					greason,
					Lusers[luserindex].Nick,
					Lusers[luserindex].Userhost);
				sprintf(sendstr, "Gline added by %s for %s [%s]",
					nick,
					ghost,
					greason);
				SendToOps(sendstr);

				/* Check if any users on the network match the new GLINE */
				for (ii = 0; ii < MAXLUSERS; ii++)
				{
					if (Lusers[ii].Userhost[0] == '\0')
						continue;
					if (WildCompare(ghost, Lusers[ii].Userhost) == 0)
					{
						sprintf(sendstr, ":%s KILL %s :%s!%s (Glined: %s)\n",
							SERVICENICK,
							Lusers[ii].Nick,
							MyName,
							SERVICENICK,
							greason);
						writesocket(Servers[CurrSIndex].socket, sendstr);	
					}	
				}
			}
			else
				notice(nick, "No privileges", sockfd);
			break;	
		} /* C_GLINE */ 
		case C_UNGLINE:
		{
			char	sendstr[MAXBUFF];
			int		ii, jj, gcnt;
			char	gname[MAXLINE];
			char	line[MAXLINE], linetemp[MAXLINE];
			char	ConfName[MAXFILELENGTH];
			char	TempFName[MAXLINE] = "t!emp667.tmp";
			char	*key;
			char	*ghost;
			FILE	*configfp;
			FILE	*fp; /* random file to store all lines except the ones */
								 /* which match the jupes 												 */

			if (CheckAccess(userindex, "g") == 1)
			{
				for (ii = 0; ii < MAXLINE; ii++)
					gname[ii] = '\0';
				GetStrIndex(gname, command, 2);
				if (gname[0] == '\0')
				{
					notice(nick, "Usage: UNGLINE <user@host>", sockfd);
					return;
				}
				strcpy(ConfName, ConfigFile);
				if ((configfp = fopen(ConfName, "r")) == NULL)
				{
					notice(nick, "Unable to open config file", sockfd);
					return;
				}
				fp = fopen(TempFName, "w"); 
				gcnt = 0;
				while (fgets(line, MAXLINE - 1, configfp))
				{
					strcpy(linetemp, line);
					key = strtok(line, ":");
					if ((strcmp(key, "g") == 0) || (strcmp(key, "G") == 0))
					{
						ghost = strtok((char *) NULL, ":");
						if (wldwld(ghost, gname) != 0)
							fputs(linetemp, fp);
						else
							gcnt++;
					}
					else
						fputs(linetemp, fp);
				} /* while () */
				fclose(configfp);
				fclose(fp); 
				(void)sprintf(sendstr, "cp %s %s", TempFName, ConfName);
				system(sendstr);
				sprintf(sendstr, "rm -f %s", TempFName);
				system(sendstr);
				for (ii = 0; ii < MAXGLINES; ii++)
				{
					if (Glines[ii].GHost[0] == '\0')
						break;
					if (wldwld(gname, Glines[ii].GHost) == 0)
						for (jj = 0; jj < MAXLINE; jj++)
							Glines[ii].GHost[jj] = '\0';
				}
				GetStrIndex(gname, command, 2);
				if (gcnt == 0)
				{
					sprintf(sendstr, "No glines matching %s found", gname);
					notice(nick, sendstr, sockfd);
					return;			
				} 
			/*	for (ii = 0; ii < MAXGLINES; ii++)
				{
					if (Glines[ii].GHost[0] == '\0')
						continue;
					if (match(
				} */
				sprintf(sendstr, "Glines for %s removed [%d matches]", gname, gcnt);
				notice(nick, sendstr, sockfd);
				if (sockfd == NODCC)
					sprintf(sendstr, "(%s!%s) !*! UNGLINE %s\n",
						Lusers[luserindex].Nick,
						Lusers[luserindex].Userhost,
						gname);
				else
					sprintf(sendstr, "#%s# ungline %s\n",
						Lusers[luserindex].Nick,
						gname);
				BroadcastDCC(-1, sendstr, DCCOPS);
				putlog(LOG1, "Glines matching %s removed by %s!%s",
					gname,
					Lusers[luserindex].Nick,
					Lusers[luserindex].Userhost);
				sprintf(sendstr, "Glines matching %s removed by %s",
					gname,
					nick);
				SendToOps(sendstr);
			}
			else
				notice(nick, "No privileges", sockfd);
			break;
		} /* C_UNGLINE */
		case C_HELP:
		{
			char	helparg[MAXLINE];
			char	sendstr[MAXLINE];
			int		ii;

			for (ii = 0; ii < MAXLINE; ii++)
				helparg[ii] = '\0';
			GetStrIndex(helparg, command, 2);
			GiveHelp(nick, helparg, sockfd);
			if (sockfd == NODCC)
				sprintf(sendstr, "(%s!%s) !*! HELP %s\n",
					Lusers[luserindex].Nick,
					Lusers[luserindex].Userhost,
					helparg);
			else
				sprintf(sendstr, "#%s# help %s\n", 
					Lusers[luserindex].Nick,
					helparg);
			BroadcastDCC(-1, sendstr, DCCOPS);
			break;	
		} /* C_HELP */
		case C_JOIN:
		{
			char	sendstr[MAXLINE];
			char	chan[MAXLINE], modes[MAXLINE];
			int		ii, OkToJoin;
			FILE	*fp;
			char	ConfFile[MAXFILELENGTH];

			if ((CheckAccess(userindex, "f") == 1) ||
					(CheckAccess(userindex, "o") == 1) ||
					(CheckAccess(userindex, "a") == 1))
			{
		#ifndef CHANNELSERVICES
				notice(nick, "Channel services are not enabled", sockfd);
				return;
		#endif

				GetStrIndex(chan, command, 2);
				if (chan[0] == '\0')
				{
					notice(nick, "Usage: JOIN <#channel> <modes>", sockfd);
					return;
				}
				GetStrIndex(modes, command, 3);
				if (modes[0] == '\0')
					strcpy(modes, DEFAULTCHANMODES);
				for (ii = 0; ii < MAXCHANS; ii++)
				{
					if (channels[ii].ChanName[0] == '\0')
						continue;
					if (strcasecmp(channels[ii].ChanName, chan) == 0)
					{
						sprintf(sendstr, "I am already monitoring %s [%s]\n", 
							channels[ii].ChanName,
							channels[ii].ChanModes);
						notice(nick, sendstr, sockfd);
						return;
					}
				}
				OkToJoin = 0;
				for (ii = 0; ii < MAXCHANS; ii++)
				{
					if (channels[ii].ChanName[0] == '\0')
					{
						OkToJoin = 1;
						break;
					}
				}
				if (!OkToJoin) /* chan limit has been reached - MAXCHANS */
				{
					notice(nick, "I am unable to monitor any more channels, maximum limit has been reached", sockfd);
					return;
				}
				join(chan);
				if (sockfd == NODCC)
					sprintf(sendstr, "(%s!%s) !*! JOIN %s [%s]\n",
						Lusers[luserindex].Nick,
						Lusers[luserindex].Userhost,
						chan,
						modes);
				else
					sprintf(sendstr, "#%s# join %s [%s]\n",
						Lusers[luserindex].Nick,
						chan,
						modes);
				BroadcastDCC(-1, sendstr, DCCOPS);
				sprintf(sendstr, "I am now monitoring %s [%s]", chan, modes);
				notice(nick, sendstr, sockfd);
				putlog(LOG1, "Now monitoring %s [%s] by request of %s!%s",
					chan,
					modes,
					Lusers[luserindex].Nick,
					Lusers[luserindex].Userhost);
				for (ii = 0; ii < MAXCHANS; ii++)
					if (channels[ii].ChanName[0] == '\0')
					{
						strcpy(channels[ii].ChanName, chan);
						strcpy(channels[ii].ChanModes, modes);
						break;
					}
				strcpy(ConfFile, ConfigFile);
				if ((fp = fopen(ConfFile, "a")) == (FILE *) NULL)
				{
					notice(nick, "Unable to open config file", sockfd);
					return;
				}
				sprintf(sendstr, "C:%s:%s\n", chan, modes);
				fputs(sendstr, fp);
				fclose(fp);
			}
			else
				notice(nick, "No privileges", sockfd);
			break;
		} /* C_JOIN */
		case C_PART:
		{
			char	sendstr[MAXLINE];
			char	pchan[MAXLINE];
			char	line[MAXLINE], linetemp[MAXLINE];
			char	*key, *ptemp;
			int		ii;
			FILE	*fp;
			FILE	*configfp;
			char	ConfName[MAXFILELENGTH];
			char	TempFName[MAXFILELENGTH] = "t!emp667.tmp";

			if ((CheckAccess(userindex, "f") == 1) ||
					(CheckAccess(userindex, "o") == 1) ||
					(CheckAccess(userindex, "a") == 1))
			{

				#ifndef CHANNELSERVICES
					notice(nick, "Channel services are not enabled", sockfd);
					return;
				#endif

				GetStrIndex(pchan, command, 2);
				if (pchan[0] == '\0')
				{
					notice(nick, "Usage: PART <#channel>", sockfd);
					return;
				}
				for (ii = 0; ii < MAXCHANS; ii++)
				{
					if (channels[ii].ChanName[0] == '\0')
						continue;
					if (strcasecmp(channels[ii].ChanName, pchan) == 0)
					{
						part(pchan);
						channels[ii].ChanName[0] = '\0';

						/* Remove channel from config file */
						strcpy(ConfName, ConfigFile);
						if ((configfp = fopen(ConfName, "r")) == NULL)
						{
							notice(nick, "Unable to open config file", sockfd);
							return;
						}
						fp = fopen(TempFName, "w"); 
						while (fgets(line, MAXLINE - 1, configfp))
						{
							strcpy(linetemp, line);
							key = strtok(line, ":");
							if ((strcmp(key, "c") == 0) || (strcmp(key, "C") == 0))
							{
								ptemp = strtok((char *) NULL, ":");
								if (strcasecmp(ptemp, pchan) != 0)
									fputs(linetemp, fp);
							}
							else
								fputs(linetemp, fp);
						} /* while () */
						fclose(fp);
						fclose(configfp);
						sprintf(sendstr, "cp %s %s", TempFName, ConfName);
						system(sendstr);
						sprintf(sendstr, "rm -f %s", TempFName);
						system(sendstr);
						if (sockfd == NODCC)
							sprintf(sendstr, "(%s!%s) !*! PART %s\n",
								Lusers[luserindex].Nick,
								Lusers[luserindex].Userhost,
								pchan);
						else
							sprintf(sendstr, "#%s# part %s\n",
								Lusers[luserindex].Nick,
								pchan);
						BroadcastDCC(-1, sendstr, DCCOPS);
						putlog(LOG1, "No longer monitoring %s by request of %s!%s",
							pchan,
							Lusers[luserindex].Nick,
							Lusers[luserindex].Userhost);
						sprintf(sendstr, "I am no longer monitoring %s", pchan);
						notice(nick, sendstr, sockfd);
						return;
					}
				}
				sprintf(sendstr, "I am not monitoring %s", pchan);
				notice(nick, sendstr, sockfd);
			}
			else
				notice(nick, "No privileges", sockfd);
			break;
		} /* C_PART */
		case C_OP:
		{
			int		ii;
			char	sendstr[MAXLINE];
			char	onick[MAXLINE];
			char	ochan[MAXLINE];

			if ((CheckAccess(userindex, "f") == 1) ||
					(CheckAccess(userindex, "o") == 1) ||
					(CheckAccess(userindex, "a") == 1))
			{
				GetStrIndex(onick, command, 2);
				GetStrIndex(ochan, command, 3);
				if (onick[0] == '\0')
					strcpy(onick, nick);
				if (ochan[0] == '\0')
					for (ii = 0; ii < MAXCHANS; ii++)
					{
						if (channels[ii].ChanName[0] == '\0')
							continue;
						sprintf(sendstr, ":%s MODE %s +o %s\n",
							SERVICENICK,
							channels[ii].ChanName,
							onick);
						writesocket(Servers[CurrSIndex].socket, sendstr);
					}
				else
				{
					if (IsChannel(ochan) != -1)
					{
						sprintf(sendstr, ":%s MODE %s +o %s\n",
							SERVICENICK,
							ochan,
							onick);
						writesocket(Servers[CurrSIndex].socket, sendstr);
					}
					else
					{
						sprintf(sendstr, "I am not monitoring %s", ochan);
						notice(nick, sendstr, sockfd);
					}
				}
				if (ochan[0] == '\0')
					strcpy(ochan, "global");
				if (sockfd == NODCC)
					sprintf(sendstr, "(%s!%s) !*! OP %s (%s)\n",
						Lusers[luserindex].Nick,
						Lusers[luserindex].Userhost,
						onick,
						ochan);
				else
					sprintf(sendstr, "#%s# op %s (%s)\n",
						Lusers[luserindex].Nick,
						onick,
						ochan);
				BroadcastDCC(-1, sendstr, DCCOPS);
			}
			else
				notice(nick, "No privileges", sockfd);
			break;
		} /* C_OP */

		default:
		{
			char 	sendstr[MAXLINE];

			if (sockfd == NODCC)
			{
				sprintf(sendstr, "[%s!%s] %s\n",
					Lusers[luserindex].Nick,
					Lusers[luserindex].Userhost,
					command);
				BroadcastDCC(-1, sendstr, DCCOPS);
			}
			notice(nick, "Unknown command", sockfd);
			break;
		} /* default */
	} /* switch (token) */
} /* ProcessCommand() */ 

/*
GetUser()
	args: char *nick
	purpose: same as IsUser()
	return: -1 if 'nick' is not a user, the index position of 'nick' in
					the array 'Users' if 'nick' is a user
*/

int GetUser(char *nick)
{
	int		ii, userindex = -1;

	for (ii = 0; ii < MAXLUSERS; ii++)
	{
		if (strcasecmp(Lusers[ii].Nick, nick) == 0)
		{
			userindex = ii;
			break;
		}	
	}
	if (userindex != -1)
		for (ii = 0; ii < MAXUSERS; ii++)
		{
			if (WildCompare(Users[ii].UserHost, Lusers[userindex].Userhost) == 0)
				return(ii);
		}
	return(-1);
}

/*
IsUser()
	args: char *nick
	purpose: determine if 'nick' matches a U: line in the config file,
					 thus being eligible to use OBServ
	return: -1 if 'nick' is not a user, the index position of 'nick' in
					the array 'Lusers' if 'nick' is a user
*/

int IsUser(char *nick)
{
	int		ii, userindex = -1;

	for (ii = 0; ii < MAXLUSERS; ii++)
	{
		if (strcasecmp(Lusers[ii].Nick, nick) == 0)
		{
			userindex = ii;
			break;
		}	
	}
	if (userindex != -1)
		for (ii = 0; ii < MAXUSERS; ii++)
			if (WildCompare(Users[ii].UserHost, Lusers[userindex].Userhost) == 0)
				return(userindex);
	return(-1);
}

/*
IsLuser()
	args: char *nick
	purpose: same as IsUser, only doesn't check if 'nick' has a U: line
	return: index position of 'nick' in array 'Lusers', -1 if 'nick'
					isn't in 'Lusers' 
*/

int IsLuser(char *nick)
{
	int		ii, userindex = -1;

	for (ii = 0; ii < MAXLUSERS; ii++)
	{
		if (strcasecmp(Lusers[ii].Nick, nick) == 0)
		{
			userindex = ii;
			break;
		}	
	}
	return(userindex);
}

/*
IsChannel()
	args: char *chan
	purpose: determine if 'chan' is currently being monitored
	return: -1 if not being monitored, index in array 'channels[]' if
					'chan' is being monitored
*/

int IsChannel(char *chan)

{
	int		ii;

	for (ii = 0; ii < MAXCHANS; ii++)
	{
		if (channels[ii].ChanName[0] == '\0')
			continue;
		if (strcasecmp(channels[ii].ChanName, chan) == 0)
			return(ii);
	}
	return(-1);
} /* IsChannel() */

int GoodTCM(char *host)

{
	int		ii;

	for (ii = 0; ii < MAXBOTS; ii++)
	{
		if (AllowedBots[ii].BotName[0] == '\0')
			break;
		printf("host=%s allowedhost=%s\n", host, AllowedBots[ii].BotHost);
		if (WildCompare(AllowedBots[ii].BotHost, host) == 0)
			return(ii);
	}
	return(-1);
} /* GoodTCM() */

int GetBot(char *bname)

{
	int		ii;

	for (ii = 0; ii < MAXDCCCONNS; ii++)
		if (connections[ii].isbot == 1)
			if (strcasecmp(connections[ii].nick, bname) == 0)
				return(ii);

	return(-1);
} /* GetBot() */

/*
IsBot()
	args: char *bname
	purpose: determine if 'bname' has a B: line 
	return: index of 'bname' in array bots[] is found, -1 if not
*/

int IsBot(char *bname)

{
	int 	ii;

	for (ii = 0; ii < MAXBOTS; ii++)
	{
		if (bots[ii].BotName[0] == '\0')
			continue;
		if (strcasecmp(bots[ii].BotName, bname) == 0)
			return(ii);
	}
	return(-1);
} /* IsBot() */

/*
IsOnDCC()
	args: char *nick
	purpose: determine if 'nick' is currently on DCC CHAT
	return: -1 if 'nick' is not on DCC CHAT, the index of 'nick' in
					array connections[] if 'nick' is on DCC CHAT
*/

int IsOnDCC(char *nick)

{
	int		ii, index = -1;

	for (ii = 0; ii < MAXDCCCONNS; ii++)
		if (strcasecmp(connections[ii].nick, nick) == 0)
		{
			index = ii;
			break;
		}

	return (index);
} /* IsOnDCC() */

/*
MatchFlags()
	args: char *Userflags, char *matchchar
	purpose: determines if 'Userflags' contains 'matchchar'
	return: 1 if there is a match, 0 if not
*/

int MatchFlags(char *Userflags, char *matchchar)
{
	int		ii;

	for (ii = 0; ii < MAXLINE; ii++)
	{
		if (Userflags[ii] == matchchar[0])
			return 1;
	}
	return 0;
}

/*
CheckAccess()
	args: int uindex, char *flag
	purpose: determines whether U: line user specified by 'uindex' has
					 the flag 'flag'.  If not, a "No privs" messages is given
					 to the user.
	return: 1 if the user has 'flag', 0 if not 				 
*/

int CheckAccess(int uindex, char *flag)
{
	char	sendstr[MAXBUFF];

	if (MatchFlags(Users[uindex].UserFlags, flag) == 1)
		return 1;
	return 0;	
}

/*
CheckJuped()
	args: char *ServerName
	purpose: check if 'ServerName' is "juped" (ie, banned from the
					 network).  If so, SQUIT it.
	return: none  
*/

void CheckJuped(char *ServerName)
{
	int		ii;
	char	sendstr[MAXBUFF];

	for (ii = 0; ii < MAXJUPES; ii++)
	{
		if (Jupes[ii].ServName[0] == '\0')
			continue;
		if (WildCompare(Jupes[ii].ServName, ServerName) == 0)
		{
			(void)sprintf(sendstr, "SQUIT %s :Juped: %s\n",
				ServerName,
				Jupes[ii].reason);
			writesocket(Servers[CurrSIndex].socket, sendstr);
			break;	
		}
	}	
} /* CheckJuped */

/*
CheckGlined()
	args: char *nick, char *userhost
	purpose: determine whether 'userhost' is GLINED from the network;
					 if so, issue a KILL for 'nick' with the reason they are
					 GLINED
	return: none
*/

void CheckGlined(char *nick, char *userhost)

{
	int		ii;
	char	sendstr[MAXBUFF];

	for (ii = 0; ii < MAXGLINES; ii++)
	{
		if (Glines[ii].GHost[0] == '\0')
			continue;
		if (WildCompare(Glines[ii].GHost, userhost) == 0)
		{
			sprintf(sendstr, ":%s KILL %s :%s!%s (Glined: %s)\n",
				SERVICENICK,
				nick,
				MyName,
				SERVICENICK,
				Glines[ii].GReason);
			writesocket(Servers[CurrSIndex].socket, sendstr);
			break;
		}	
	}
} /* CheckGlined */

/*
WildCompare()
	args: char *wildexp, char *regstr
	purpose: determine if 'regstr' (which contains NO wildcards) matches
					 possibly wild carded string 'wildexp'
	return: 0 if they match, 1 if they do not match

	*** NOTE: This function was coded by Hendrix (jimi@texas.net) and
						is used under the terms of the GNU General Public License
*/

int WildCompare(char *wildexp, char *regstr)
{
	while (*wildexp == '*' || tolower(*wildexp) == tolower(*regstr) || 
				 *wildexp == '?')
		if (*wildexp == '*')
		{
			while (*(++wildexp) == '*' || (*wildexp == '?' && *(wildexp + 1) == '*'));
			if (*wildexp)
			{
				while (*regstr)
					if (!WildCompare(wildexp, regstr++))
						return 0;
				return 1;
			}
			else
				return 0;
		}
		else
			if (!*wildexp)
				return 0;
			else
				if (!*regstr)
					return 1;
				else
				{
					++wildexp;
					++regstr;		
				}
	return 1;				
}

/*
wldwld()
	args: char *wild1, char *wild2
	purpose: attempt to determine if wild carded strings 'wild1' and
					 'wild2' match.
	return: 0 if they match, 1 if they do not match

	(This function is much less efficient than WildCompare() because
	 it is possible that it will return a 0 even if the two strings
	 don't match at all.  For example, this function will make a 
	 successful match between *@*.com and *@*.net.*, even though the 
	 two are not a very good match.  So, use only when necessary.)

	*** NOTE: This function was coded by Hendrix (jimi@texas.net) and
						is used under the terms of the GNU General Public License
*/

int wldwld (wild1,wild2)
char *wild1,*wild2;
{
  while (tolower(*wild1) == tolower(*wild2) || *wild1 == '*' || *wild2 == '*' ||
         *wild1 == '?' || *wild2 == '?') {
    if (*wild1 == '*') {
      /* This will stop idiots who ban *?*?*?*?*?*?* from crashing the bot */
      while (*(++wild1) == '*' || (*wild1 == '?' && *(wild1+1) == '*'));
      if (*wild1) {
        while (*wild2)
          if (!wldwld(wild1,wild2++))
            return 0;
        return 1;
        }
      else
        return 0;
      }
    else if (*wild2 == '*') {
      /* This will stop idiots who ban *?*?*?*?*?*?* from crashing the bot */
      while (*(++wild2) == '*' || (*wild2 == '?' && *(wild2+1) == '*'));
      if (*wild2) {
        while (*wild1)
          if (!wldwld(wild1++,wild2))
            return 0;
        return 1;
        }
      else
        return 0;
      }
    else if (!*wild1)
      return (*wild2 == '?');
    else if (!*wild2)   /* wild1 must be a ? in this case */
      return 1;
    else {
      ++wild1;
      ++wild2;
      }
    }
    return 1;
}

/*
match()
	args: char *mask, char *name
	purpose: attempt to determine if 'name' matches 'mask'.  Both
					 may contain wildcards.
	return: 0 if match, 1 if not

	(This function is much more efficient than wldwld().)

	** NOTE: This function was taken from standard ircd 2.8.x source.
*/

int match(char *mask,char *name)
{
  unsigned char *m = (u_char *)mask, *n = (u_char *)name;
  char  *ma = mask, *na = name;
  int   wild = 0, q = 0;

  while (1)
    {
      if (*m == '*')
        {
          while (*m == '*')
            m++;
          wild = 1;
          ma = (char *)m;
          na = (char *)n;
        }

      if (!*m)
        {
          if (!*n)
            return 0;
          for (m--; (m > (u_char *)mask) && (*m == '?'); m--)
            ;
          if ((*m == '*') && (m > (u_char *)mask) &&
              (m[-1] != '\\'))
            return 0;
          if (!wild)
            return 1;
          m = (u_char *)ma;
          n = (u_char *)++na;
        }
      else if (!*n)
        {
          while(*m == '*')
            m++;
          return (*m != 0);
        }
      if ((*m == '\\') && ((m[1] == '*') || (m[1] == '?')))
        {
          m++;
          q = 1;
        }
      else
        q = 0;

      if ((tolower(*m) != tolower(*n)) && ((*m != '?') || q))
        {
          if (!wild)
            return 1;
          m = (u_char *)ma;
          n = (u_char *)++na;
        }
      else
        {
          if (*m)
            m++;
          if (*n)
            n++;
        }
    }
}
